/*
   Author:  Jianfei Zhu  
            Concordia University
   Date:    Sep. 26, 2003

Copyright (c) 2003, Concordia University, Montreal, Canada
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

   - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
   - Neither the name of Concordia University nor the names of its
       contributors may be used to endorse or promote products derived from
       this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.

*/

/* This is an implementation of FP-growth* / FPmax* /FPclose algorithm.
 *
 * last updated Sep. 26, 2003
 *
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <time.h>
#include "common.h"
#include "buffer.h"
#include "hash_table.h"
//#include "fp_tree.h"
#include "fk_tree.h"
#include "tree.h"


#define LINT sizeof(int)

int* ITlen;        //the number of itemset with a specific length
int* bran;         //the number of nodes in each level
int* prefix;       //just used in enumerate itemset

int* order_item;   // given order i, order_item[i] gives the original itemname, corresponding to table
int* item_order;   // given item i, item_order[i] gives its new order, corresponding to order
				   // order_item[item_order[i]]=i; item_order[order_item[i]]=i;

int* order_support;

int* fullItem=NULL;     // the items with full support, maybe item, maybe the index.
int  fullNum=0;

bool* current_fi;   //to be use for maximal or close patterns.
int*  compact;      //the current itemset, represented by the set of orders in current table.
int*  supp;         //use in conditional_pattern_base  to store support temporarily

stack* list;
hash_table *patternTable;

int TRANSACTION_NO=0;
int POSITIVE_NO=0;
int ITEM_NO=100;
int THRESHOLD;
double RATIO_THRESHOLD=0;
int TIDLENGTH = 0;

memory* fp_buf;
memory* fp_buf1;


//for debug
int nodeCount=0;

double printLen()
{
	int i, j;
	double total=0;
	for(i=ITEM_NO-1; i>=0&&ITlen[i]==0; i--); //find the lengthy frequent itemset
	for(j=0; j<=i; j++) //print from length 0 to length maximal.
    {
		printf("%d\n", ITlen[j]);
		total+=ITlen[j];
    }
	printf("total: %.0f\n",total);

	return total;
}


void main(int argc, char **argv)
{
	//bool   bClass=false;
	//int    posItem=-1;
	char *inFileName;
	char *outFileName;
	double threshold;
	//RATIO_THRESHOLD = 0;
	int    nFull=0;
	int i=1, index =0;

	if (argc < 4)
	{
		cout << "usage: ecgen <infile> <MINSUP> <outfile>\n";
		exit(1);
	}

    while(i<argc)
    {

		switch(index++)
		{
         case 0:
			 inFileName=argv[i];
			 break;
         case 1:
			 threshold = atof(argv[i]);
			 THRESHOLD = threshold;
			 break;
         case 2:
			 outFileName=argv[i];
			 break;
		}
		i++;
    }

	clock_t startTime=clock();
	

	FK_tree* fptree;

	Data* fdat=new Data(inFileName);

	if(!fdat->isOpen()) {
		cerr << inFileName << " could not be opened!" << endl;
		exit(2);
	}

	fp_buf=new memory(60, 4194304L, 8388608L, 2);
	fp_buf1=new memory(60, 4194304L, 8388608L, 2);

    fptree = (FK_tree*)fp_buf->newbuf(1, sizeof(FK_tree));   
	fptree->bClass=false;
	fptree->posItem=-1;

	fptree->init(-1, 0);
	fptree -> scan1_DB(fdat, threshold);                  //will be different for FK and FI
	ITlen = new int[fptree->itemno];
	bran = new int[fptree->itemno];
	compact = new int[fptree->itemno];
	prefix = new int[fptree->itemno];


	list=new stack(fptree->itemno, true);  //should change 
	patternTable=new hash_table(fptree->itemno, ITEM_NO);
	ITEM_NO=fptree->itemno;
	//TIDLENGTH = fptree->head[0]->count;

	assert(list!=NULL && bran!=NULL && compact!=NULL && ITlen!=NULL && prefix!=NULL);

	for(i =0; i < fptree->itemno; i++)
	{
		ITlen[i] = 0L;
		bran[i] = 0;
	}

	fptree->scan2_DB(fdat);
    fdat->close();

	//fptree->traverseTree(fptree->Root, 0);

	FSout* fout=NULL;
	FSout *foutClose=NULL;
	if(index>=3)
	{
		char tempFileName[128];
		strcpy(tempFileName,outFileName);
		strcat(tempFileName,".close");
		foutClose = new FSout(tempFileName);
		if(fptree->fullSize>0)
		 //if(testRatio(TRANSACTION_NO,POSITIVE_NO,TRANSACTION_NO,POSITIVE_NO,RATIO_THRESHOLD))
		{
         	int * fullSet = new int[fptree->fullSize];
			int clen = 0;
			Copy(fullSet, clen, fptree->fullSet, fptree->fullSize);
			patternTable->insert(fullSet,
				fptree->fullSize, TRANSACTION_NO, POSITIVE_NO, NULL, 0);

			//foutClose->printSet(fptree->fullSize,fptree->fullSet,TRANSACTION_NO, POSITIVE_NO);
			//foutClose->printString("\n");
			nFull=1;
		 }

		
	}


	if(fptree->Single_path())
	{

		Knode* node;

		int i=0;
		for(node=fptree->Root->leftchild; node!=NULL; node=node->leftchild)
		{
			    list->FS[i++]=node->itemname;

				if(node->count==TRANSACTION_NO)
                  i--;
				else
				  list->counts[i-1] = node->count;  // put more to the stack
		}


		fptree->generate_all(fptree->itemno, fout,foutClose); //top is 0 at this moment. //should change
		
	}

	current_fi = new bool[fptree->itemno];
	supp=new int[fptree->itemno];		//for keeping support of items
	assert(supp!=NULL&&current_fi!=NULL);

	for(i = 0; i<fptree->itemno; i++)
	{
		current_fi[i] = false;
		supp[i]=0; //initialize before usage, and clear after usage.
	}

	double closePatterns=nFull;
	double keyPatterns=nFull;

    fptree->FKP_growth(fout,foutClose);

	//fptree->insertTid();
	
	if(fout)
	{
		fout->close();
		delete fout;
	}
	
	T_tree* ttree = (T_tree*) fp_buf1->newbuf(1, sizeof(T_tree));
	ttree->init();

	ttree->scan_FP(fptree);

	delete fp_buf; //added to avoid memory leak

	fout = new FSout(outFileName);
	
	foutClose->printDigit(TRANSACTION_NO);
	foutClose->printString("\n");
	foutClose->printDigit(ITEM_NO);
	foutClose->printString("\n");
	foutClose->printOrderItem();
	patternTable->outputClose1(foutClose,keyPatterns, closePatterns);
	//patternTable->outputClose(foutClose,keyPatterns, closePatterns, ttree);
		 //foutClose->printTime("Total Time",(clock()-startTime)*1.0/CLOCKS_PER_SEC);
	ttree->outputTree(fout);
	foutClose->close();
	fout->close();

	delete foutClose;
	delete fout;	 
	delete patternTable;


	double totalPatterns=printLen();
	
	delete fdat;
	//delete fp_buf; //added to avoid memory leak
	delete fp_buf1;
	delete []ITlen;
	delete []compact;
	delete []bran;
	delete []prefix;

	delete list; //stack
	delete []current_fi;
	delete []supp;

	delete []order_item;
	delete []item_order;	
}
