#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <iostream.h>
#include "buffer.h"
#include "common.h"
#include "fk_tree.h"
#include "hash_table.h"

int *testSet=NULL;
int *subSet=NULL;
int *localSupp=NULL;
int *globalSupp=NULL;

static int KNODEPTRSIZE=sizeof(Knode*);
static int KNODESIZE=sizeof(Knode);
static int FKTREESIZE=sizeof(FK_tree);

bool FK_tree::bClass= false;
int  FK_tree::posItem=-1;
bool Rprune = false;
int RpruneDepth = 0;

extern int* order_support;
extern hash_table *patternTable;


//////////////////////////////////////////////////////
//	Internal functions --- sorting
//////////////////////////////////////////////////////

#define SORTHRESH 9

template <class T> void swap(T* k, T* j)
{
	T temp;
	temp = *j;
	*j = *k;
	*k = temp;
}

int findpivot(const int& i, const int& j) {return (i+j)/2;}

int partition(int* array, int* temp, int l, int r, int pivot)
{
	do {
		while (array[++l] > pivot);
		while (r && array[--r] < pivot);
		swap(array+l, array+r);
		swap(temp+l, temp+r);
	}while (l < r);
	swap(array + l, array + r);
	swap(temp + l, temp + r);
	return l;
}


//Bug! Dec. 4, 2003
void inssort(int* array, int* temp, int i, int j)
{
	for (int k=i+1; k<=j; k++)
		for (int m=k; (m>i) && (array[m] > array[m-1]); m--)
		{
			swap(array+m, array+m-1);
		    swap(temp+m, temp+m-1);
		}
}


void sort(int *array, int* temp, int i, int j)
{
	if(j-i < SORTHRESH) inssort(array, temp, i, j);
	else
	{
		int pivotindex = findpivot(i, j);
		swap(array+pivotindex, array+j);
		swap(temp+pivotindex, temp+j);
		int k = partition(array, temp, i-1, j, array[j]);
		swap(array+k, array+j);
		swap(temp+k, temp+j);
		if((k-i)>1) sort(array, temp, i, k-1);
		if((j-k)>1) sort(array, temp, k+1, j);
	}
}

/////////////////////////////////////////////////
//  class stack
/////////////////////////////////////////////////

stack::stack(int length, bool close)
{
	top= 0; 
	FS = new int[length];
	if(close)
		counts = new int[length];   //should change here
	else
		counts = NULL;
}


stack::~stack()
{
	delete []FS;

	if(counts) //added for memory leak
		delete counts;
}

/////////////////////////////////////
////// Extended by Li Haiquan
void FK_tree::init(int old_itemno, int new_itemno)
{
	int i;
  	Root = (Knode*)fp_buf->newbuf(1, KNODESIZE);
	Root->init(NULL, -1, 0, 0);
	if(old_itemno!=-1)
	{
		order = (int*)fp_buf->newbuf(ITEM_NO, INTSIZE);
		count = (int*)fp_buf->newbuf(old_itemno, INTSIZE);
		table = (int*)fp_buf->newbuf(old_itemno, INTSIZE);
		for (i=0; i<old_itemno; i++)
		{
			order[i]=-1;
			count[i] = 0;
			table[i] = i;
		}
		for (; i<ITEM_NO; i++)
			order[i]=-1;

	    if(list->top==0)
		   //cout<<"stack is empty, no prefix exist!"<<endl;
		   printf("stack is empty, no prefix exist!\n");
	   else
           prefixCount=list->counts[list->top-1-new_itemno];  //error here, 
	}      

	itemno = new_itemno;
	if(new_itemno!=0)
		head = (Knode**)fp_buf->newbuf(itemno, KNODEPTRSIZE);

	fullSet=NULL;
	fullSize=0;
	prefixSet=NULL;
	prefixSize=0;
}
  
FK_tree::~FK_tree()
{
}

void FK_tree::scan1_DB(Data* fdat, double threshold)
{
	int i,j;
	int net_itemno=0; //current maximal item number
	int *counts;      // original item num--> counts

	counts= new int[ITEM_NO];
	assert(counts!=NULL);

	for(i=0; i<ITEM_NO; i++)
		counts[i] = 0;

	Transaction *Tran = new Transaction;
	assert(Tran!=NULL);
  
	int start=0;
	if(bClass) start=1;

	while(Tran = fdat->getNextTransaction(Tran))
	{	
		for(int i=start; i<Tran->length; i++) 
		{
			if(Tran->t[i]>=ITEM_NO)
			{
				ITEM_NO = 2*Tran->t[i];
				int* temp = new int[2*Tran->t[i]];
				for(j=0; j<=net_itemno; j++)
					temp[j] = counts[j];
				for(; j<ITEM_NO; j++)
					temp[j] = 0;
				delete []counts;
				counts=temp;
				net_itemno=Tran->t[i];
			}else
				if(net_itemno<Tran->t[i])
					net_itemno = Tran->t[i];
			counts[Tran->t[i]]++;
		}
		TRANSACTION_NO++;
	}
 
    if(threshold<1)
      THRESHOLD=ceil(threshold*TRANSACTION_NO);
	else
	  THRESHOLD=ceil(threshold);

    prefixCount=TRANSACTION_NO;

	ITEM_NO = net_itemno+1;

//	cout << "The dataset contains " << TRANSACTION_NO << " transactions and "
//       << "the largest item is " << net_itemno << endl;

//	THRESHOLD = ceil (TRANSACTION_NO/100000.0 * THRESHOLD);
//	cout << "The minimum support is "<<THRESHOLD<<endl;

	order = (int*)fp_buf->newbuf(ITEM_NO, INTSIZE);
	table = (int*)fp_buf->newbuf(ITEM_NO, INTSIZE);
	count = (int*)fp_buf->newbuf(ITEM_NO, INTSIZE);
	for (i=0; i<ITEM_NO; i++)
	{
		order[i]=-1;
		if(counts[i]!=prefixCount)
		  count[i] = counts[i];  
		else
        {
		  count[i] = -1*counts[i]; //replace the counts if equal to maximal 
		  fullSize++;
        }
		//should remember all the items.
		table[i] = i;
	}
   
	sort(count, table, 0, ITEM_NO-1);
 	
	for (i =0; i<ITEM_NO&&count[i] >= THRESHOLD; i++);
 
	itemno = i;   // thrink the real frequent items
    if(fullSize>0)
	{ 
		//fullSet=new int[fullSize];
		fullSet = (int*)fp_buf->newbuf(fullSize, INTSIZE); //changed for memory leak
		for (i =ITEM_NO-1; i>=0&&count[i]<0; i--)
		{
			fullSet[ITEM_NO-1-i]=i;
		}
    }//set the fullSet here.
	for (j=0; j<itemno; j++)
	{
		count[j]=counts[table[j]];  
		order[table[j]]=j;
	}

//	for(int k=0; k<ITEM_NO; k++)cout<<order[k]<<"  "; cout<<endl;
//	for( k=0; k<itemno; k++)cout<<table[k]<<"  "; cout<<endl;

	head =(Knode**)(fp_buf->newbuf(itemno, KNODEPTRSIZE));

	if(itemno>SUDDEN+5)
	{
		//array = new int*[itemno-1-SUDDEN];
		array = (int**)fp_buf->newbuf(itemno-1-SUDDEN,sizeof(int*));//for memory leak
		for(i=0; i<itemno-1-SUDDEN; i++)
		{
			//array[i] = new int[itemno-1-i];
			array[i] = (int*)fp_buf->newbuf(itemno-1-i,INTSIZE); //for memory leak
			for(j=0; j<itemno-1-i; j++)
				array[i][j] = 0;
		}
	}else array = NULL;

// The following is for the case when ITEM_NO is very big and fptree->itemno is very small
	order_item = new int[ITEM_NO];
	item_order = new int[ITEM_NO];
	for(i=0; i<itemno; i++)
	{
		head[i] = (Knode*)fp_buf->newbuf(1, KNODESIZE);
		((Knode *)head[i])->init(NULL, i, count[i], 0);
		//head[i]->TidSetSize = count[i];
		//head[i]->TidSet = (int *)fp_buf->newbuf(head[i]->TidSetSize, sizeof(int));

		order_item[i]=table[i]; // original item name, from order -->original item
		table[i]=i;  //transfer the name
		item_order[i] = order[i]; //from original item-->order
		order[i]=i;
	}
	for(;i<ITEM_NO; i++)
	{
		item_order[i] = order[i]; //from original item-->order, for other items.
		order[i]=-1; 
		if(count[i]<0)
			order_item[i]=table[i];
		else
			order_item[i]=-1; //remove infrequent items
	}
	//ITEM_NO = itemno;
	//for debug
	/*int k;
	order_support = new int[ITEM_NO];
	for(k=0;k<ITEM_NO;k++)
	{
		if(order_item[k]>=0)
		{
			//order_support[k] = count[k];
			printf("%d ",order_item[k]);
			printf("(%d)",count[k]);
		}
	}

	scanf("test");*/

	delete []counts;
	delete Tran;
    
	testSet=(int *)fp_buf->newbuf(itemno,INTSIZE);
	subSet=(int *)fp_buf->newbuf(itemno,INTSIZE);
	localSupp=(int *)fp_buf->newbuf(itemno,INTSIZE);
	globalSupp=(int *)fp_buf->newbuf(itemno,INTSIZE);
	for(i=0;i<itemno;i++)
    {
       localSupp[i]=0;
	   globalSupp[i]=0;
    }
}

int FK_tree::conditional_pattern_base(Knode* node,int currPrefixCount,int *currKeySet, int currKeySize)
{
	Knode* temp, *parent;
	int i;

	for(i=0; i<order[node->itemname]; i++)
		supp[i]=0;  //it is okay here.

	for(temp=node->next; temp!=NULL; temp=temp->next)
	{
		parent=temp->par;
		for(; parent->itemname!=-1;parent=parent->par)
			supp[order[parent->itemname]]+=temp->count;
	}

	int k=0;
	int listlen=list->top;
	for(i=0; i<order[node->itemname]; i++)
	{
		if(supp[i]>=currPrefixCount) //?will > error
        {
           supp[i]=-1*supp[i];   //change here.
        }else if(supp[i]>=THRESHOLD)   //should change here.
		{
	        int len=0;
			Copy(testSet,len,currKeySet,currKeySize); //should copy from prefixSet
			Insert(testSet,len,table[i]);
			bool bKey=keyCheck(testSet,len,supp[i],table[i]);
			if(bKey)
            {
			  k++;
			  list->FS[list->top++]=table[i];
			  list->counts[list->top-1] = supp[i];//changed.
            }
			else
              supp[i]=-1*supp[i];  //should change here, reverse the value.
		}else
			supp[i] = 0; 
	}
    //delete iSet;
	return k;
}

int FK_tree::conditional_pattern_base(int itemname,int currPrefixCount,int *currKeySet, int currKeySize)
{
	int i, k=0, item = itemno-1-order[itemname];
	int listlen=list->top;

	for(i=itemno-2-item; i>=0;i--)
	{
		if(array[item][i]==currPrefixCount)
        {
	       supp[itemno-2-item-i]=-1*array[item][i];
        }
		else if(array[item][i]>=THRESHOLD)   //should change here.
		{
  			int len=0;
	        Copy(testSet,len,currKeySet,currKeySize); 
			Insert(testSet,len,table[itemno-2-item-i]);
			//bool bKey = false; //for debug only
			bool bKey=keyCheck(testSet,len,array[item][i],table[itemno-2-item-i]);
            if(bKey)
            {
			  k++;
			  list->FS[list->top++]=table[itemno-2-item-i];
			  supp[itemno-2-item-i]=array[item][i];
			  list->counts[list->top-1] = array[item][i]; //should change here.
			} //otherwise, by default, it is zero
			else
 			  supp[itemno-2-item-i]=-1*array[item][i];
		}
		else
 			  supp[itemno-2-item-i]=0;  //added in Dec 3
		//otherwise, record the itemset
	}

	return k;
}


void FK_tree::scan1_DB(FK_tree* old_tree,int *keySet, int keySize)
{
	int i, j;
	int* old_order = old_tree->order;
       //find the full set first, copy down from the old tree
	for(i=0; i< itemno; i++)
	{
		count[i]=supp[old_order[list->FS[i+list->top-itemno]]];
		table[i]=list->FS[i+list->top-itemno];
		supp[old_order[list->FS[i+list->top-itemno]]]=0; //clear the space
	}

	sort(count, table, 0, itemno-1);

	for(i=0; i<itemno; i++)
	{
		order[table[i]]=i;
		head[i] = (Knode*)fp_buf->newbuf(1, KNODESIZE);
		head[i]->init(NULL, table[i], count[i], 0);
		//head[i]->TidSetSize = count[i];
		//head[i]->TidSet = (int*) fp_buf->newbuf(head[i]->TidSetSize, sizeof(int));
	}

	if(itemno > SUDDEN+5 && old_tree->level_25 > SWITCH)
	{
		array = (int**)fp_buf->newbuf(itemno-1-SUDDEN, sizeof(int*));
		for(i=0; i<itemno-1-SUDDEN; i++)
		{
			array[i] = (int*)fp_buf->newbuf(itemno-1-i, INTSIZE);
			for(j=0; j<itemno-1-i; j++)
				array[i][j] = 0;
		}
	}else array = NULL;

	prefixSet=keySet;
	prefixSize=keySize;
}

#ifdef CLOSE
//get tailSet the support array, will allocate the memory
bool FK_tree::GetTailSet(int support, int *&sup, int start, int end, int *&retSet, int &retSize)
{   
   retSet=(int*)fp_buf->newbuf(end-start+1,INTSIZE);
   retSize=0;
   for (int i=start; i<=end;i++)
   {
     if (sup[i]==support)
		retSet[retSize++]=i;
	 sup[i]=0;
   } 
  return true;
}

bool FK_tree::CheckSubtree(Knode *node, int *localSupp, int &start, int &end, int *sumSupp)
{
  start=node->itemname;
  end=node->itemname;
  int i;
  
  Knode *curr=node;
  while(curr!=NULL)
  {
    localSupp[curr->itemname]+=curr->count; 
    if(curr->itemname<start)
	  start=curr->itemname;
    if(curr->itemname>end)
      end=curr->itemname;
	bool bStop=false;
    

    int support=curr->tailSupport;
    Knode *temp=curr->leftchild;
    while(temp!=NULL)
	{
        support=support+temp->count;
	    temp=temp->rightsibling;
	}
	if(support<curr->count)
      bStop=true;
	if(!bStop)
	{
      for(int i=0;i<curr->tailSize;i++)
		 {
	       localSupp[curr->tailSet[i]]+=curr->tailSupport; 
	       if(curr->tailSet[i]<start)
		     start=curr->tailSet[i];
	       if(curr->tailSet[i]>end)
		    end=curr->tailSet[i];
	  }//should change this part.
	}//end handling nonempty tail
	
	if(curr->leftchild==NULL)
	  bStop=true;
	if(!bStop)
	{
	  curr=curr->leftchild;
      continue;
	}
    if(curr==node)
	  break;
   	if(curr->rightsibling!=NULL)
	{
	   curr=curr->rightsibling;
	   continue;
    }
    //finish the current tree;
    while((curr->rightsibling==NULL)&&(curr!=node))
	  curr=curr->par;
    if(curr==node)
      break;
    curr=curr->rightsibling;
  }//end while
  localSupp[node->itemname]=0; 
  sumSupp[node->itemname]=0;
  for(i=start;i<=end;i++)
  {
     if(localSupp[i]!=node->count)
     {
       start++;
	   localSupp[i]=0;
	   sumSupp[i]=0;
     }else 
	   break;
  }
  for(i=end;i>=start;i--)
  {
     if(localSupp[i]!=node->count)
     {
       end--;
	   localSupp[i]=0;
	   sumSupp[i]=0;
     }else
	   break;
  }
  for(i=start; i<=end;i++)
  {
     if(localSupp[i]==node->count)
       sumSupp[i]+=localSupp[i];  
  }
  return true;
}
#endif

/*
bool FK_tree::CheckSubtree(Knode *node, int *localSupp, int &start, int &end, int *sumSupp)
{
  //if(node->tailSize==0)
 //	  return false;
  Knode *temp=node->leftchild;
  int support=0;
  while(temp!=NULL)
  {
     support=support+temp->count;
	 localSupp[temp->itemname]+=temp->count; 
	 sumSupp[temp->itemname]+=temp->count;
	 if(temp->itemname<start)
		 start=temp->itemname;
	 if(temp->itemname>end)
		 end=temp->itemname;
	 temp=temp->rightsibling;
  }
  if(node->tailSize>0)
  {
    int diffSupp=node->count-support;
    for(int i=0;i<node->tailSize;i++)
    {
	 localSupp[node->tailSet[i]]+=diffSupp; 
	 sumSupp[node->tailSet[i]]+=diffSupp;
	 if(node->tailSet[i]<start)
		 start=node->tailSet[i];
	 if(node->tailSet[i]>end)
		 end=node->tailSet[i];
	}
  }else if(support<node->count)  //???
	  return false;
  temp=node->leftchild;
  while(temp!=NULL)
  {
     CheckSubtree(temp,localSupp,start,end,sumSupp);
	 temp=temp->rightsibling;
  }

  return true;
}
*/

bool FK_tree::ScanNodeLink(FK_tree *old_tree, Knode *node, int * &retFullSet, int &retFullSize, int bInsert)
{
    int i, has, current; //has: length of frequnt sub-path, current: the current node order
	int* origin=NULL;  //itemset of itemname based on the new order table
	Knode* link, *parent;
	int*  old_order;
	int * old_table;
	int lMin,lMax,gMin, gMax;

	old_order = old_tree->order;
	old_table=  old_tree->table;

	current = old_order[node->itemname];
    
	if(bInsert)
    {
	  origin = new int[current];
	  assert(origin!=NULL);

	  for(i=0; i<current; i++) origin[i]=-1;
	}

	gMin=0;
	gMax=ITEM_NO-1;	//bug???? fixed by mengling 08/01/05
    lMin=0;
	lMax=0;
    
	//inefficient way
    //for(i=0;i<ITEM_NO;i++)
	//	globalSupp[i]=0;

	for(link=node->next; link!=NULL; link=link->next)
	{	
#ifdef CLOSE
		CheckSubtree(link, localSupp,lMin,lMax,globalSupp);
		if(lMin>gMin) 
        {
           int end=lMin;
		   if(lMin>gMax)
			   end=gMax+1;
           for(i=gMin;i<end;i++)
			   globalSupp[i]=0;
			gMin=lMin;
		}
		else
         for(i=lMin;i<gMin;i++)
			 globalSupp[i]=0;  //a bug here, found in Dec 2, 2004.

		if(lMax<gMax) 
        {
           int end=lMax;
           if(gMin>lMax)
              end=gMin-1;
	       for(i=gMax;i>end;i--)
			   globalSupp[i]=0;		//bug???? will fall out of bound
			gMax=lMax;
        }else
          for(i=gMax+1;i<=lMax;i++)
			  globalSupp[i]=0;
#endif
		has=0;
		if(bInsert)
        {
		  for(parent=link->par; parent->itemname!=-1; parent=parent->par)   
			if(order[parent->itemname]!=-1)
			{
				origin[order[parent->itemname]]=parent->itemname;
				has++;
			}
			else
              if(supp[old_order[parent->itemname]]<0)
               if((node->count+supp[old_order[parent->itemname]])>0)
               {
                  localSupp[parent->itemname]+=link->count;
				  if(parent->itemname<lMin)
					  lMin=parent->itemname;
				  if(parent->itemname>lMax)
					 lMax=parent->itemname; 
			   } //added in Dec 3th.
          
			//else should check support from supp, insert into tail
		  if(has)
		  {
			fill_count(origin, link->count);
			/*TID_STRUCT tidstt;
			tidstt.set = true;
			tidstt.tidLength = link->TidLength;
			tidstt.tidSet = link->TidSet;*/
			//Knode *lastNode=insert(compact, link->count, has, link->posCount, &tidstt);
			Knode *lastNode=insert(compact, link->count, has, link->posCount);
			
#ifdef CLOSE
            int *tailSet=NULL;
			int tailSize=0;
			if(lMin<=lMax)
			  GetTailSet(link->count,localSupp,lMin,lMax,tailSet,tailSize);
            if(tailSize>0)
		 	  lastNode->addTail(tailSet,tailSize,link->count);
#endif
		  }else
            if(lMin<=lMax)
		      for(i=lMin;i<=lMax;i++)				   
		        localSupp[i]=0; 
        }else
		{
	      if(lMin<=lMax)
		    for(i=lMin;i<=lMax;i++)				   
		      localSupp[i]=0;
		  if(gMin>gMax)
			  break;
		}

	}
    if(bInsert)
    {
	  cal_level_25(); //update the index
	  delete []origin;
    }
	int *tailSet=NULL;
 	int tailSize=0;
#ifdef CLOSE
	if(gMin<=gMax)
    {
	   int support=old_tree->head[current]->count;
	   for(i=gMin;i<=gMax;i++)
       {
		   if(globalSupp[i]!=support)
           {
			   globalSupp[i]=0;
			   gMin++;
           }
		   else
			 break;
	   }
       for(i=gMax;i>=gMin;i--)
       {
		   if(globalSupp[i]!=support)
           {
			   globalSupp[i]=0;
			   gMax--;
           }
		   else
			 break;
	   }
       if(gMin<=gMax)
		 GetTailSet(support,globalSupp,gMin,gMax,tailSet,tailSize);
	}
#endif
	int newFullNum=0;
	for(i=0; i<current;i++)  // a bug here previously,found in Dec 2th, 04
    {
		if((supp[i]+node->count)==0)
			newFullNum++;  //should check the value here.
    } //may have problem here.
	retFullSize=old_tree->fullSize+newFullNum+tailSize;
	if(retFullSize>0)
    {
 	  retFullSet=(int *)fp_buf->newbuf(retFullSize,INTSIZE);
	  Copy(retFullSet,retFullSize,old_tree->fullSet,old_tree->fullSize);
      for(i=0; i<current;i++) //another bug here previously. found in Dec 2th, 04
	  {
		if((supp[i]+node->count)==0)
        {
		  Insert(retFullSet,retFullSize, old_table[i]); //an error here.
		  supp[i]=0; //should also check here.
        }
	  }  //should record table[i];
	}
    for(i=0;i<tailSize;i++)
    {
		Insert(retFullSet,retFullSize,tailSet[i]); //is it neccessary to order?
    }
  

	return true;
}

void FK_tree::scan2_DB(FK_tree* old_tree, Knode* node)
{
    ScanNodeLink(old_tree,node,fullSet, fullSize,true);
	
} 

Knode *FK_tree::insert(int* compact, int counts, int current, int posCount)
{
	Knode* child = (Knode *)Root;
	Knode* temp, *temp1=NULL;
	int i=0;

	while(i<current)
	{
		for(temp=child->leftchild; temp!=NULL; temp =temp->rightsibling)
		{
			if(temp->itemname==table[compact[i]])break;
			if(temp->rightsibling==NULL)temp1 = temp;
		}

		if(!temp)break;
		
		//found a match
		temp->count+=counts;
		temp->posCount+=posCount;
		//temp add tid
		
		head[order[temp->itemname]]->posCount+=posCount;
		//head table add tid as well

		child=temp;
		i++;
	}
	//for not matched ones
	while(i<current)
	{   
		child = child->append(this, temp1, table[compact[i]], counts, posCount);
		bran[i]++;
		i++;
	}

	return child;
}

Knode* FK_tree::insert(int* compact, int counts, int current, int posCount, TID_STRUCT* tidstt)
{
	Knode* child = (Knode *)Root;
	Knode* temp, *temp1=NULL;
	int i=0;

	while(i<current)
	{
		for(temp=child->leftchild; temp!=NULL; temp =temp->rightsibling)
		{
			if(temp->itemname==table[compact[i]])break;
			if(temp->rightsibling==NULL)temp1 = temp;
		}

		if(!temp)break;
		
		//found a match
		temp->count+=counts;
		temp->posCount+=posCount;
		
		//temp add tid (only add tid at the last node)
		//if(tidstt->set)
		//	temp->addTidSet(tidstt->tidSet, tidstt->tidLength);
		//else
		//	temp->addTid(tidstt->tid);

		head[order[temp->itemname]]->posCount+=posCount;
		//head table add tid as well
		//if(tidstt->set)
		//	head[order[temp->itemname]]->addTidSet(tidstt->tidSet, tidstt->tidLength);
		//else
		//	head[order[temp->itemname]]->addTid(tidstt->tid);

		child=temp;
		i++;
	}
	//for not matched ones
	while(i<current)
	{   
		child = child->append(this, temp1, table[compact[i]], counts, posCount);
		/*child->TidSetSize = count[compact[i]];
		child->TidSet = (int*)fp_buf->newbuf(child->TidSetSize, sizeof(int));
		if(tidstt->set)
			child->addTidSet(tidstt->tidSet, tidstt->tidLength);
		else
			child->addTid(tidstt->tid);*/
		
		bran[i]++;
		i++;
	}

	if(child->TidSet)
	{
		if(tidstt->set)
			child->addTidSet(tidstt->tidSet, tidstt->tidLength);
		else
			child->addTid(tidstt->tid);
	}
	else
	{
		child->TidSetSize = count[compact[--i]];
		child->TidSet = (int*)fp_buf->newbuf(child->TidSetSize, sizeof(int));
		if(tidstt->set)
			child->addTidSet(tidstt->tidSet, tidstt->tidLength);
		else
			child->addTid(tidstt->tid);
	}

	return child;
}

void FK_tree::scan2_DB(Data *fdat)
{
	int i, j, has; //has: number of frequent patterns in a transaction
	int* origin, *buffer=new int;  //a vector of orders in current head table for frequent items.
	Transaction *Tran = new Transaction;
	origin = new int[ITEM_NO];
	assert(Tran!=NULL&&origin!=NULL && buffer!=NULL);

	for(j=0; j<ITEM_NO; j++) origin[j]=-1;
    int start=0;
	if(bClass) start=1;

	for(i=0; i<TRANSACTION_NO; i++)
	{
		Tran = fdat->getNextTransaction(Tran);
		has=0;
		for(int j=start; j<Tran->length; j++) 
		{
			if(item_order[Tran->t[j]]!=-1)
			{
				has++;
				if(origin[item_order[Tran->t[j]]]!=-1)
					has--;         // in transaction, items are not supposed to appear more than twice
				origin[item_order[Tran->t[j]]] = item_order[Tran->t[j]];
			}
		}
		int posCount=1;
		if(bClass)
		  if(Tran->t[0]!=posItem)
			posCount=0;
        POSITIVE_NO+=posCount;
		
		if(has){
			
			fill_count(origin, 1);
			//need to fill in tid as well
			TID_STRUCT tidstt;
			tidstt.set = false;
			tidstt.tid = i;
			
		
			
			//insert(compact, 1, has, posCount)
				insert(compact, 1, has, posCount, &tidstt); //no need to set anything  //maybe 1 or zero
		}
	}

	cal_level_25();

	delete []origin;
	delete buffer;
	delete Tran;
}

bool FK_tree::printStack()
{
  for(int i=0;i<list->top;i++)
  {
    //cout<<list->FS[i]<<":"<<list->counts[i]<<" ";
	printf("%d:%d ",list->FS[i],list->counts[i]);
  }
  //cout<<endl;
  printf("\n");
  return true;
}

void FK_tree::generate_all(int new_item_no, FSout* fout, FSout *foutClose) //generate all set for a single path, the number is in first parameter, items are in list
{
 Knode* node;
 int newFullNum=0;
 for(node=Root->leftchild; node!=NULL; node=node->leftchild)
   {
	  int currCount=this->head[order[node->itemname]]->count;
	  if(currCount>=prefixCount)
      {
		  printStack();
		  //cout<<"error in generate single path key pattern for abnormal count"<<endl;
		  printf("error in generate single path key pattern for abnormal count");
		  continue;
	  } //will not be true
      list->FS[list->top++] = node->itemname;
	  ITlen[list->top-1]++;
	  //should give a sorted set for the key
	  int *currKeySet=new int[prefixSize+1];
	  int  currKeySize;
	  Copy(currKeySet,currKeySize,prefixSet,prefixSize);
      Insert(currKeySet,currKeySize,node->itemname);
	  outputKey(fout,currKeySize, currKeySet, currCount); //should update here.
#ifdef CLOSE
	  newFullNum++;
	  //if(testRatio(currCount,node->posCount,TRANSACTION_NO,POSITIVE_NO,RATIO_THRESHOLD))
      if(true)
	  {
		  int *tailSet=NULL;
		  int tailSize=0;
		  int tempFullSize=0;
		  int start;
		  int end;
		  CheckSubtree(node, localSupp,start,end,globalSupp);
		  if(start<=end)
		  {
			GetTailSet(node->count,localSupp,start,end,tailSet,tailSize);
			for(int i=start;i<=end;i++)
				globalSupp[i]=0;
		  }
		  
		  tempFullSize=list->top+fullSize+newFullNum+tailSize;
		  int *tempFullSet=new int[tempFullSize];
 		  if(tempFullSize>0)
		  {
 			Copy(tempFullSet,tempFullSize,fullSet,fullSize);
			Knode *temp=node->par;
			while(temp!=NULL)
			{
			  if(temp->itemname<0)
				  break;
			  Insert(tempFullSet,tempFullSize,temp->itemname);
			  temp=temp->par;
			}
  		  } 
		  for(int i=0;i<tailSize;i++)
		  {
			 Insert(tempFullSet,tempFullSize,tailSet[i]);
		  }
		  for(i=0;i<list->top;i++)
			 Insert(tempFullSet,tempFullSize,list->FS[i]);
		  
		  int index = patternTable->insert(tempFullSet,tempFullSize,currCount,node->posCount,currKeySet,currKeySize);
		  //patternTable->insertTidSet(index, node->TidSet, node->TidLength);
	  }
#endif
	  list->top--;
   }
}

bool  FK_tree::pruneByRatio(Knode* node)
{
   Knode* temp;
   int minCount=node->count-node->posCount;   
   for(temp=node->next; temp!=NULL; temp=temp->next)
   {
     int negSup=temp->count-temp->posCount;
	 if(negSup<minCount)
		 minCount=negSup;
   }
  bool bLarge=testRatio(node->posCount+minCount,node->posCount,TRANSACTION_NO,POSITIVE_NO,RATIO_THRESHOLD);

  return bLarge==false;
}

int FK_tree::FKP_growth(FSout* fout, FSout *foutClose)
{
	int sequence, current, new_item_no, listlen; 
	   //sequence: the item sequence to be handled one by one
	   //current: the current transformed itemname
	   //new_item_no: the number of frequent items in conditional base
	   //listlen: the saved stack length
	int MC=0;			//markcount for memory
	unsigned int MR=0;	//markrest for memory
	char* MB;			//markbuf for memory

	Knode* Current;   //the current handling node in the head table

	for(sequence=0; sequence<itemno; sequence++)
	{
		Current=head[sequence];
		current=head[sequence]->itemname;
		//if(current==7)
		//	bool debug = true;
   		list->FS[list->top++]=head[sequence]->itemname;
		list->counts[list->top-1]=head[sequence]->count; //must add here.
		listlen = list->top;
		
		/*if(!Rprune) //debugged by mengling
		{
			if(pruneByRatio(Current))
			{
				//list->top = listlen-1; //finish the current one
				Rprune = true;
				RpruneDepth = listlen;
				//int x=0;
				//continue;
			}
		}*/
  		
		
		ITlen[list->top-1]++;
        int *currKeySet=new int[prefixSize+1];
		int currKeySize=0;
		Copy(currKeySet,currKeySize,prefixSet,prefixSize);
		Insert(currKeySet,currKeySize,head[sequence]->itemname);
		outputKey(fout, currKeySize, currKeySet, this->head[sequence]->count);
	
		if(array && sequence>SUDDEN+1)  //should change
			//new_item_no = 0; //for debug only
			new_item_no=conditional_pattern_base(Current->itemname,head[sequence]->count,currKeySet,currKeySize);  //new_item_no is the number of elements in new header table. put the item and support in stack and supp
		else if(sequence !=0)   //should change
			new_item_no=conditional_pattern_base(Current,head[sequence]->count,currKeySet,currKeySize);  //new_item_no is the number of elements in new header table.
		else
			new_item_no = 0;  //the top node.
		
	
	//	if(new_item_no==0 || new_item_no == 1)
		if(new_item_no==0 )
		{   /*
			if(new_item_no==1)    
			{
				ITlen[list->top-1]++;
				outputKey(fout,list->top, list->FS, supp[order[list->FS[list->top-1]]]);
				supp[order[list->FS[list->top-1]]] = 0; //clear the support
			} //one more key pattern. //may be an error here. 
			*/ //just regard it as a tree;

			//should output close pattern here. also the should traversed.
#ifdef CLOSE
		    
			if(!Rprune)
			{
				//if(testRatio(this->head[sequence]->count,this->head[sequence]->posCount,TRANSACTION_NO,POSITIVE_NO,RATIO_THRESHOLD))
				if(true)
				{
					
					int *tempFullSet=NULL;
					int tempFullSize=0;
					ScanNodeLink(this,Current,tempFullSet,tempFullSize,false);
					int *closeSet=new int[tempFullSize+listlen];
					int  cLen=0;
					Copy(closeSet,cLen,tempFullSet,tempFullSize);
					for(int i=0;i<listlen;i++)
						Insert(closeSet,cLen,list->FS[i]);
					int index = patternTable->insert(closeSet,cLen,this->head[sequence]->count,this->head[sequence]->posCount,currKeySet,currKeySize);
					//patternTable->insertTidSet(index, this->head[sequence]->TidSet, this->head[sequence]->TidLength);
				}
			}
			//outputClose(foutClose,list->FS,listlen,tempFullSet,tempFullSize,this->head[sequence]->count);
#endif

			//delete tempFullSet;
			if(listlen==RpruneDepth)
			{
				RpruneDepth = 0;
				Rprune = false;
			}
			list->top=listlen-1; //pop up the current item
			//should output here.
			continue;
		} //in case single path, we should output the fullSet.
        
		//may still need scan the downstreams.

		FK_tree *fptree;  //generate a tree for each of item
		MB=fp_buf->bufmark(&MR, &MC); //mark the current position

		fptree = (FK_tree*)fp_buf->newbuf(1, FKTREESIZE);
		fptree->init(this->itemno, new_item_no);

		fptree->scan1_DB(this,currKeySet, currKeySize); //use the stack here.
		fptree->scan2_DB(this, Current);
        
		//scan downstream and also summarize all the tail items.

		list->top=listlen;	//restore the top to the prefix itemset.	

#ifdef CLOSE
		if(!Rprune)
		{
			// output the close patterns here.
			//if(testRatio(this->head[sequence]->count,this->head[sequence]->posCount, TRANSACTION_NO,POSITIVE_NO,RATIO_THRESHOLD))
			if(true)
			{
				int *closeSet=new int[fptree->fullSize+listlen];
				int  cLen=0;
				Copy(closeSet,cLen,fptree->fullSet,fptree->fullSize);
				for(int i=0;i<listlen;i++)
					Insert(closeSet,cLen,list->FS[i]);
				int index = patternTable->insert(closeSet,cLen, this->head[sequence]->count,this->head[sequence]->posCount,currKeySet,currKeySize);
				//patternTable->insertTidSet(index, this->head[sequence]->TidSet, this->head[sequence]->TidLength);
			}
		}
		
#endif
	
		if(fptree->Single_path())
		{
                               /* patch   Oct. 9, 2003*/
            list->top = listlen;  //remove some code here.
                              /*************************/
			fptree->generate_all(new_item_no, fout,foutClose); //should change here.
			list->top--; //finish the current one
		}else{             //not single path
		    fptree->FKP_growth(fout, foutClose);
			list->top = listlen-1; //finish the current one
		} //pay attention here also
		fp_buf->freebuf(MR, MC, MB);
		
		if(listlen==RpruneDepth)
		{
				RpruneDepth = 0;
				Rprune = false;
		}
	}

	return 0;
}

bool FK_tree::keyCheck(int *iSet, int length, int support, int newItem)
{
  int i;
  if(length<=2)
	  return true;

  int currSupport=0;
  for(i=1;i<length;i++)
   subSet[i-1]=iSet[i];
  for(i=0;i<length;i++)
  {
    if(i>0)
      subSet[i-1]=iSet[i-1];
	if(iSet[i]==newItem)
		continue;//skip the last one 
	//int index = 0; //for debug only
	int index=patternTable->find(subSet,length-1, currSupport);
	if(index<0)
  		return false;
  	if(support>=currSupport)
		return false;
  }
  return true;
} //should not copy the itemset.

bool FK_tree::outputKey(FSout* fout,int length,int *iSet, int support)
{
  patternTable->insert(iSet,length,support); //should copy the set
  
  if(fout)
  {
	  fout->printSet(length, iSet, support);
  }
  return true;
}

#ifdef CLOSE
bool FK_tree::outputClose(FSout *fout, int *iKeySet, int keyLen, int *inFullSet, int inFullSize, int support)
{
  if(fout)
  {
	  fout->printset(inFullSize, inFullSet);
	  fout->printSet(keyLen, iKeySet,support);
  }
  return true;
}

#endif	

void FK_tree::cal_level_25()
{
	int i, total_25=0, total_50=0, total_bran=0, maxlen=0;

	for(i=0; i<this->itemno && bran[i]!=0; i++);
	maxlen =i;
	for(i=0; i<int(maxlen*0.25); i++)
		total_25 +=bran[i];
	total_50 = total_25;
	for(i=int(maxlen*0.25); i<this->itemno*0.5; i++)
		total_50 +=bran[i];
	for(i=0; i<this->itemno && bran[i]!=0; i++)
	{
//		cout<<i<<" " <<bran[i]<<endl;
		total_bran+=bran[i];
		bran[i]=0;
	}
	level_25 = (double)total_25/total_bran*100;
//	cout<<"First 25% levels: "<<(double)total_25/total_bran*100<<"%  "<<"50% levels: "<<(double)total_50/total_bran*100<<"%"<<endl;
}

void FK_tree::fill_count(int* origin, int support)
{
	int i, j=0;
	for(i=0; i<itemno; i++)
	{
		if(origin[i]!=-1)
		{
			compact[j++]=i; //save all the frequent items from the trans in frequency order
			origin[i] = -1;
		}
	}

	if(array)
	{
		int comp_len = j;
		for(i=comp_len-1; i>0 && compact[i]>SUDDEN; i--)
			for(j=i-1; j>=0; j--)
				array[itemno-1-compact[i]][compact[i]-compact[j]-1]+=support;
	}
}


bool FK_tree::Single_path(bool close)const
{
	Knode* node;

	for(node=Root->leftchild; node!=NULL; node=node->leftchild)
		if(node->rightsibling!=NULL)return false;

	if(close)
		for(node=Root->leftchild; node!=NULL; node=node->leftchild)
		{
			list->FS[list->top] = node->itemname;  //maybe a bug here.
			list->counts[list->top++] = node->count;
		}

	return true;
}

inline bool testRatio(int totalSupport, int positiveSupport, int dbSize, int posSize, double threshold)
{
  if(threshold==0)
	  return true;

  int negSize=dbSize-posSize;
  int negativeSupport=totalSupport-positiveSupport;
  
  int falsePositive=posSize-positiveSupport;
  int trueNegative=negSize-negativeSupport;
  if((negativeSupport==0)||(falsePositive==0))
  {
	  if((positiveSupport==0)||(trueNegative==0))
		  return 1>=threshold; //changed by mengling
	  return true;
  }
  
  int  denominator=positiveSupport*trueNegative;
  int  numerator=negativeSupport*falsePositive;
  double ratio=denominator;
  ratio=ratio/numerator;
  if(ratio>=threshold)
	  return true;
  return false;
}

bool FK_tree::checkPath(Knode* node, int* iset, int length)
{	
	if(node->depth < length)
		return false;

	int i=length-2;
	node = node->par;

	while(node!=NULL 
		&& node->itemname >=0 
		&& i>=0){

		if(node->depth < i+1)
			return false;

		if(node->itemname == iset[i])
			i--;

		node = node->par;
	}
	
	if(i>=0)
		return false;

	return true;
}

bool FK_tree::traverseTree(Knode* node, int level)
{
	node = node->leftchild;
	level++;
	bool check;

	if(node==NULL)
		return true;

	do{

		if(level!=node->depth)
			return false;

		check = traverseTree(node, level);
		
		if(check = false)
			return false;

		node = node->rightsibling;
	}
	while(node!=NULL);

	return true;
}

bool mergeSets(int* dstSet, int &dstLen, int* srcSet, int srcLen)
{
	if(dstSet == NULL || srcSet == NULL)
	{
		printf("merge error!");
		scanf("press to exit");
		exit(2);
	}
	
	for(int i=0; i<srcLen; i++)
	{
		Insert(dstSet, dstLen, srcSet[i]);
	}

	return true;
}

int FK_tree::getTidSet(Knode* node, int* tidSet, int &tidLen)
{
	if(node==NULL || tidSet == NULL)
		return 0;
	
	Knode* current = node->leftchild;
	
	if(node->TidSet && node->TidLength >0)
		mergeSets(tidSet, tidLen, node->TidSet, node->TidLength);

	while(current)
	{
		getTidSet(current, tidSet, tidLen);
		current = current->rightsibling;
	}
	
	return 1;
}

/*int FK_tree::insertTid()
{
	int i;
	int closeSize = patternTable->getCloseSize();
	LPCloseHashNode* closeTable = patternTable->getCloseTable();

	
	for(i=0; i<closeSize;i++)
	{
		LPCloseHashNode lpCloseNode=closeTable[i];
		if(lpCloseNode==NULL)
			continue;

		int* iset = lpCloseNode->iset;
		int length = lpCloseNode->length;
		
		for(; length>0; length--)
		{
			if(order[iset[length-1]]>=0)
				break;
		}

		if(length<1)
		{
			lpCloseNode->TidLength = TRANSACTION_NO;
			lpCloseNode->TidSet = NULL;
			continue;
		}

		//creat tid set
		int MaxLen = lpCloseNode->support;
		lpCloseNode->TidSet = (int*) fp_buf->newbuf(MaxLen, sizeof(int));
		lpCloseNode->TidLength = 0;

		Knode* node = head[iset[length-1]]->next;
		while(node!=NULL){
			
			if(checkPath(node, iset, length))
			{
				//mergeSets(lpCloseNode->TidSet, lpCloseNode->TidLength, node->TidSet, node->TidLength);
				getTidSet(node, lpCloseNode->TidSet, lpCloseNode->TidLength);

				if(lpCloseNode->TidLength>MaxLen)
				{
					printf("error merging tids");
					scanf("exit");
					exit(2);
				}
			}
			node = node->next;
		}
	}
	
	return 1;
}*/

int FK_tree::scanTid(int* iset, int length, int*& tidSet, int maxLen, int &tidLen)
{
	if(iset==NULL || tidSet == NULL)
		return 0;
	
	for(; length>0; length--)
	{
			if(order[iset[length-1]]>=0)
				break;
	}

	if(length<1)
	{
			tidLen = TRANSACTION_NO;
			tidSet = NULL;
			return 1;
	}

	
	Knode* node = head[iset[length-1]]->next;
	while(node!=NULL){
		
		if(checkPath(node, iset, length))
		{
			//mergeSets(tidSet, tidLen, node->TidSet, node->TidLength);
			getTidSet(node, tidSet, tidLen);

			if(tidLen>maxLen)
			{
				printf("error merging tids");
				scanf("exit");
				exit(2);
			}
		}
		node = node->next;
	}

	return 1;
}