// hash_table.cpp: implementation of the hash_table class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include "hash_table.h"
#include "math.h"
#include "common.h"
#include "tree.h"
#include "FK_tree.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
#define MAXTRY 5
#define INCREMENT(x) x+x+x
#define MAXTIMES 512

extern bool isSubset(int* controlSet, int conlen, int* caseSet, int caselen);

hash_table::hash_table(int freqNum, int totalNum)
{
   keySize=0;
   closeSize=0;
   keyThreshold=0;
   closeThreshold=0;
   currKeyNum=0;
   currCloseNum=0;
   mapSize=0;
   keyMap=NULL;
   closeMap=NULL;

   //secModMap=NULL;
   keyTable=NULL;
   closeTable=NULL;

   int initSize=INITSIZE;
   //int estSize=TRANSACTION_NO*itemNum*1024;
   //if(estSize<initSize)
   //initSize=getPrime(INITSIZE);

   //cout<<"hash table size is "<<initSize<<endl;
   printf("hash table size is %d\n",initSize);
   //int initSize=getPrime(TRANSACTION_NO*2);
   initKeyTable(initSize);
   initCloseTable(initSize);

   initModMap(freqNum,keyMap,initSize);
   initModMap(totalNum,closeMap,initSize);

   //for debug
   numInsert=0;
   numProbe=0;
 }

hash_table::~hash_table()
{
   releaseTable();
   if(keyMap!=NULL)
	   delete keyMap;
   if(closeMap!=NULL)
	   delete closeMap;
   if(numInsert!=0)
	   //cout<<"insert "<<numInsert<<" probe "<<numProbe<<" average "<<numProbe*1.0/numInsert<<endl;
		printf("insert %d probe %d average %f\n",numInsert,numProbe, numProbe*1.0/numInsert);
}

bool hash_table::initKeyTable(int newSize)
{
  keyTable=new LPKeyHashNode[newSize];
  for( int i=0;i<newSize;i++)
  {
     keyTable[i]=NULL;
  }
  keySize=newSize;
  currKeyNum=0;
 
  keyThreshold=LOADFACTOR*keySize;
 
  return true;
}

bool hash_table::initCloseTable(int newSize)
{
	closeTable=new LPCloseHashNode[newSize];
	for( int i=0;i<newSize;i++)
	{
		closeTable[i]=NULL; 
	}
	closeSize=newSize;
	currCloseNum=0;
	
	closeThreshold=LOADFACTOR*closeSize;
	
	return true;
}

bool hash_table::releaseTable()
{
  for(int i=0;i<keySize;i++)
  {
    if(keyTable[i]!=NULL)
    {
      delete keyTable[i]->iset;
      delete keyTable[i];
	  keyTable[i]=NULL;
	  currKeyNum--;
    }
  }
  for(i=0;i<closeSize;i++)
  {
    if(closeTable[i]!=NULL)
    {
	   /*KeySetNode *node=closeTable[i]->head;
	   while(node!=NULL)
	   {
          KeySetNode *temp=node;
		  node=node->next;
		  //delete []temp->iset; //debugged by mengling for memory link
		  delete temp;
	   }*/
	   delete closeTable[i]->iset;  
	   delete closeTable[i];
	   currCloseNum--;
    }
  }
  delete keyTable;
  delete closeTable;
  keyTable=NULL;
  closeTable=NULL;
  keySize=0;
  closeSize=0;
  return true;
}

int  hash_table::single_mod(int i, int prime)
{
	int ret=-1;
	if(i<0) return -1;
	unsigned long pValue=0;
	if(i<MAXPOWER)
		ret=(unsigned long)(pow(2,i)) % prime;
	else
		ret=(2*single_mod(i-1, prime)) % prime;
	
	if(ret<0)
	{
		//cout<<"single mod return negative"<<ret<<endl;
		printf("single mod return negative %d", ret);
	}
	return ret;
}

int  hash_table::set_mod(int *iset, int length, int order, int *map, int size)
{
	int sum=0;
	
	for(int i=0;i<length;i++)
	{
		//if(order==0)
		if(iset[i]>=0) //ignore removed transactions
			sum=sum+map[iset[i]];
		while(sum>=size)
			sum=sum-size;
		//else
		//     sum=sum+secModMap[iset[i]];
	}
	//if(order==0)
	//else
	//  ret=sum % (size+OFFSET);
	
	return sum;
}

bool hash_table::compareKey(int *iset1, int length1, int *iset2, int length2)
{
	if(length1!=length2)
		return false;
	for(int i=length1-1;i>=0;i--)
	{
		if(iset1[i]!=iset2[i]) return false;
	}
	return true;
}

int hash_table::insert(int *iset, int length, int support)
{
  int i=0;
  int initIndex=0;
  int index;
  
  initIndex=set_mod(iset,length,0,keyMap,keySize);
  index=initIndex;
  while(i<keySize)
  {
  
	if(keyTable[index]==NULL)
    {
	    LPKeyHashNode lpNode=new KeyHashNode;
		lpNode->iset=iset;
		lpNode->length=length;
		lpNode->support=support;
		keyTable[index]=lpNode;
		currKeyNum++;
		if(currKeyNum>keyThreshold)
        {
			//for debug
			//cout<<"now rehash key table from "<<currKeyNum<<endl;
			printf("now rehash key table from %d\n",currKeyNum);
			rehashKeyTable();
        }
        return index;
	}
    i++;
  	index=initIndex+INCREMENT(i);  
    while(index>=keySize)
	{
      index=index-keySize;
	}

  }
  return -1;
}



int   hash_table::insert(int * & iCloseSet, int cLen, int support, int posCount, int *iKeySet, int kLen)
{
	int i=0;
	int initIndex=0;
	int index;
	initIndex=set_mod(iCloseSet,cLen,0,closeMap, closeSize);
	index=initIndex;
	
	while(i<closeSize)
	{
		if(index == 4265939)
		{
			bool debug = true;
		}
		
		if(closeTable[index]==NULL)
		{
			LPCloseHashNode lpNode=new CloseHashNode;
			lpNode->iset=iCloseSet;
			lpNode->length=cLen;
			lpNode->support=support;
			//lpNode->posCount=posCount;
			//lpNode->head=lpKeyNode;
			lpNode->keys = (K_tree*)fp_buf1->newbuf(1, sizeof(K_tree));
			lpNode->keys->init();
			lpNode->keys->insertKey(iKeySet, kLen);
			
			closeTable[index]=lpNode;
			currCloseNum++;
			if(currCloseNum>closeThreshold)
			{
				printf("now rehash close hash table from %d\n", currCloseNum);	
				rehashCloseTable();
			}
			return index;
		}else
		{
			LPCloseHashNode lpNode=closeTable[index];
			bool bEqual=compareKey(iCloseSet,cLen, lpNode->iset, lpNode->length);
			if(bEqual)
			{
				//lpKeyNode->next=lpNode->head;
				//lpNode->head=lpKeyNode;
				lpNode->keys->insertKey(iKeySet, kLen);
				delete iCloseSet; //added by mengling for memory leak
				return index;
			}    
		}
		i++;
		index=initIndex+INCREMENT(i);  
		while(index>=closeSize)
		{
			index=index-closeSize;
		}
		
	}
	
	return -1;
}

int  hash_table::insert(LPCloseHashNode lpCloseNode)
{
	int i=0;
	int initIndex=0;
	int index;
	
	initIndex=set_mod(lpCloseNode->iset,lpCloseNode->length,0, closeMap,closeSize);
	index=initIndex;
	
	while(i<closeSize)
	{
		if(closeTable[index]==NULL)
		{
			currCloseNum++;
			closeTable[index]=lpCloseNode;
			if(currCloseNum>closeThreshold)
			{
				//for debug
				//cout<<"now rehash close table from "<<currCloseNum<<endl;
				printf("now rehash close table from %d\n",currCloseNum);
				rehashCloseTable();
			}
			return index;
		}else
		{
			LPCloseHashNode lpCurrNode=closeTable[index];
			bool bEqual=compareKey(lpCloseNode->iset,lpCloseNode->length,lpCurrNode->iset, lpCurrNode->length);
			if(bEqual)
			{
			/*KeySetNode *lpKeyNode=lpCurrNode->head;
			while(lpKeyNode->next!=NULL)
			lpKeyNode=lpKeyNode->next;
				lpKeyNode->next=lpCloseNode->head;*/
				printf("rehash problem!");
				scanf("exit!");
				return index;
			}    
		}
		i++;
		index=initIndex+INCREMENT(i);  
		while(index>=closeSize)
		{
			index=index-closeSize;
		}
		
	}
	
	return -1;
}

int  hash_table::find(int *iset, int length,int & support)
{
  int i=0;
  int initIndex=0;
  int index;
  
  initIndex=set_mod(iset,length,0,keyMap,keySize);
  index=initIndex;
  while(i<keySize)
  {
  	LPKeyHashNode lpNode=keyTable[index];
	if(lpNode!=NULL)
    {
		//bool bEqual = true; //for debug only
		bool bEqual=compareKey(iset,length, lpNode->iset, lpNode->length);
		if(bEqual)
        {
          support=lpNode->support;
            //for debug
          numInsert++;
          numProbe+=i+1;
          return index;
        }
	}
	else
    {
      //for debug
      numInsert++;
      numProbe+=i+1;
	  return -1;
    }
    i++;
    //int index=getProbe(iset, length, i);   	
	index=initIndex+INCREMENT(i);  
    while(index>=keySize)
	{
        index=index-keySize;
	}	
  }
  return -1;
}

bool hash_table::rehashKeyTable()
{
  LPKeyHashNode *oldTable=keyTable;
  int oldSize=keySize;
  keySize=getPrime(2*keySize);
  initKeyTable(keySize);
  initModMap(mapSize,keyMap,keySize);
  for(int i=0;i<oldSize;i++)
  {
    if(oldTable[i]!=NULL)
      insert(oldTable[i]->iset,oldTable[i]->length,oldTable[i]->support);
  }
  delete oldTable;
  return true;
}

bool hash_table::rehashCloseTable()
{
	LPCloseHashNode *oldTable=closeTable;
	int oldSize=closeSize;
	closeSize=getPrime(2*closeSize);
	initCloseTable(closeSize);
	initModMap(mapSize,closeMap,closeSize);
	for(int i=0;i<oldSize;i++)
	{
		if(oldTable[i]==NULL)
			continue;
		insert(oldTable[i]);  
	}
	delete oldTable;
	return true;
}

int  hash_table::getPrime(int start)
{
	int iCurr=start;
	while(1)
	{
		if(isPrime(iCurr))
			return iCurr;
		iCurr++;
	}
	
	return -1;
}

bool hash_table::isPrime(int num)
{
	if(num<=2) return false;
	
	for(int i=2;i<(num/2);i++)
		if((num %i)==0) return false;
		
		return true;
}


bool  hash_table::initModMap(int num, int * & map, int prime)
{
	if(mapSize>0)
	{
		if(map!=NULL)
			delete map;
		//if(secModMap!=NULL)
		//	delete secModMap;
	}
	map=new int[num];
	//secModMap=new int[num];
	mapSize=num;
	for(int i=0;i<num;i++)
		map[i]=single_mod(i,prime);
	
	return true;
}

bool  hash_table::begin()
{
	curr=0;
	
	return true;
}

bool  hash_table::next(int * & iset, int & length, int &support)
{
  for(; curr<keySize;curr++)
  {
    if(keyTable[curr]!=NULL)
    {
      iset=keyTable[curr]->iset;
	  length=keyTable[curr]->length;
	  support=keyTable[curr]->support;
	  curr++;
      return true;
    }
  }
  return false;
}

bool  hash_table::outputClose(FSout *fout, double & keyNum, double &closeNum, T_tree* ttree)
{
	for(int i=0; i<closeSize;i++)
	{
		LPCloseHashNode lpCloseNode=closeTable[i];
		if(lpCloseNode==NULL || lpCloseNode->support == TRANSACTION_NO)
			continue;
		
		//output support
		int *tidset = new int[lpCloseNode->support];
		int support = 0;
		support = ttree->outputTidSet(tidset, lpCloseNode->tids, lpCloseNode->support);
		//support = lpCloseNode->support;
		
		if(support < THRESHOLD)
		{
			delete []tidset;
			continue;
		}
		
		//output closed pattern
		fout->printOrderSet(lpCloseNode->length, lpCloseNode->iset, 0);
		
		fout->printString("(");
		fout->printDigit(support);
		fout->printString(")");
		
		//output key patterns
		K_tree* ktree = lpCloseNode->keys;
		if(ktree)
			ktree->outputKeys(fout, lpCloseNode->length);
		
		fout->printString("|");
		
		//output tidlist
		fout->printTid(&tidset[lpCloseNode->support-support], support);
		
		delete []tidset;
		//fout->printTid(lpCloseNode->TidSet, lpCloseNode->TidLength);
		//fout->printTime("Tid:",lpCloseNode->TidLength);
		//fout->printString("\n");
		closeNum++;
	}
	return true;
}

bool  hash_table::outputClose1(FSout *fout, double & keyNum, double &closeNum, T_tree* ttree)
{
	for(int i=0; i<closeSize;i++)
	{
		LPCloseHashNode lpCloseNode=closeTable[i];
		if(lpCloseNode==NULL)
			continue;
		
		if(lpCloseNode->support < THRESHOLD)
			continue;
		//output index
		fout->printString("[");
		fout->printDigit(i);
		fout->printString("]");

		//output closed pattern
		fout->printOrderSet(lpCloseNode->length, lpCloseNode->iset, 0);
		
		fout->printString("(");
		fout->printDigit(lpCloseNode->support);
		fout->printString(")");
		
		//output key patterns
		K_tree* ktree = lpCloseNode->keys;
		if(ktree)
			ktree->outputKeys(fout, lpCloseNode->length);

		//output tidlist
		if(lpCloseNode->support!=TRANSACTION_NO && ttree)
		{
			int *tidset = new int[lpCloseNode->support];
			ttree->outputTidSet(tidset, lpCloseNode->tids, lpCloseNode->support);
			
			fout->printString("|");
			fout->printTid(tidset, lpCloseNode->support);
			
			delete []tidset;
		}
		else
			fout->printString("\n");
		
		closeNum++;
	}
	return true;
}

int hash_table::insertEC(int *&iCloseSet, int clen, int support, K_tree *keys)
{
	if(iCloseSet == NULL)
		return 0;
	
	int i=0;
	int initIndex=0;
	int index;
	initIndex=set_mod(iCloseSet,clen,0,closeMap, closeSize);
	index=initIndex;
	
	while(i<closeSize)
	{
		if(closeTable[index]==NULL)
		{
			LPCloseHashNode lpNode=new CloseHashNode;
			lpNode->iset=iCloseSet;
			lpNode->length=clen;
			lpNode->support=support;
			lpNode->keys = keys;
			lpNode->tids = NULL;
			closeTable[index]=lpNode;
			currCloseNum++;
			if(currCloseNum>closeThreshold)
			{
				//for debug
				printf("now rehash close hash table from %d\n", currCloseNum);
				
				rehashCloseTable();
			}
			return index;
		}
		
		i++;
		index=initIndex+INCREMENT(i);  
		while(index>=closeSize)
		{
			index=index-closeSize;
		}
		
	}
	
	return 0;
}

int hash_table::insertTidPointer(int index, Tnode* tidPt)
{
	if(tidPt == NULL || closeTable[index]==NULL)
		return 0;
	
	LPCloseHashNode lpNode = closeTable[index];
	lpNode->tids = tidPt;
	
	return index;
}

int hash_table::mergeECs(int index1, int index2)
{
	LPCloseHashNode lpCloseNode1=closeTable[index1];
	LPCloseHashNode lpCloseNode2=closeTable[index2];
	
	if(lpCloseNode1 == NULL || lpCloseNode2 == NULL)
		return -255;
	
	if(lpCloseNode1->length>lpCloseNode2->length)
	{
		//merge keys
		K_tree *temp = lpCloseNode1->keys;
		lpCloseNode1->keys = lpCloseNode2->keys;
		lpCloseNode1->keys->mergeTree(temp, lpCloseNode1->length);
		
		return index1;
	}
	else
	{
		//merge keys
		K_tree *temp = lpCloseNode2->keys;
		lpCloseNode2->keys = lpCloseNode2->keys;
		lpCloseNode2->keys->mergeTree(temp, lpCloseNode2->length);
		return index2;
	}
}
/*
int hash_table::insertTidSet(int index, int *tidSet, int tidLength)
{
LPCloseHashNode lpCloseNode=closeTable[index];

  if(tidSet == NULL || lpCloseNode == NULL)
		return 0;
		
		  lpCloseNode->TidLength = tidLength;
		  lpCloseNode->TidSet = tidSet;
		  
			return 1;
			}
			
			  int hash_table::insertTid(int *&tidSet, int tidLen, int *&iCloseSet, int cLen, int support)
			  {
			  int i=0;
			  int initIndex=0;
			  int index;
			  initIndex=set_mod(tidSet,tidLen,0,closeMap, closeSize);
			  index=initIndex;
			  
				while(i<closeSize)
				{
				if(closeTable[index]==NULL)
				{
				LPCloseHashNode lpNode=new CloseHashNode;
				lpNode->TidSet = tidSet;
				lpNode->TidLength = tidLen;
				lpNode->iset=iCloseSet;
				lpNode->length=cLen;
				lpNode->support=support;
				lpNode->head = NULL;
				closeTable[index]=lpNode;
				currCloseNum++;
				if(currCloseNum>closeThreshold)
				{
				//for debug
				//cout<<"now rehash close hash table from "<<currCloseNum<<endl;
				printf("now rehash close hash table from %d\n", currCloseNum);
				
				  rehashCloseTable();
				  }
				  return index;
				  }else
				  {
				  LPCloseHashNode lpNode=closeTable[index];
				  //check support
				  if(lpNode->support == tidLen)
				  {
				  //check subset superset relationship
				  bool test =false;
				  bool update = false;
				  
					if(cLen > lpNode->length)
					{
					test = isSubset(iCloseSet, cLen, lpNode->iset, lpNode->length);
					update = true;
					}
					else
					test = isSubset(lpNode->iset, lpNode->length, iCloseSet, cLen);
					
					  if(test)//merge
					  {
					  if(update)
					  {
					  lpNode->iset = iCloseSet;
					  lpNode->length = cLen;
					  }
					  else
					  {
					  }
					  }
					  
						}
						}
						i++;
						index=initIndex+INCREMENT(i);  
						while(index>=closeSize)
						{
						index=index-closeSize;
						}
						
						  }
						  
							return -1;
}*/