#include <assert.h>
#include "tree.h"
#include "hash_table.h"
#include "buffer.h"
#include "FK_tree.h"

//////////////////////////////////////////////////////////////////////
// Functions of T_tree Class
//////////////////////////////////////////////////////////////////////

extern memory* fp_buf1;
extern int TRANSACTION_NO;
extern int ITEM_NO;
extern int* order_item;
extern hash_table *patternTable;
extern int* fullItem;
extern int fullNum;

//////////////////////////////////////////////////////////////////////
// external variables
//////////////////////////////////////////////////////////////////////

void K_tree::init()
{
	Root = (K_node*)fp_buf1->newbuf(1, sizeof(K_node));
	Root->init(NULL, -1);

	header = NULL;
}

K_tree::~K_tree()
{
}

K_node* K_tree::insertKey(int *keySet, int len)
{
	if(keySet==NULL)
		return NULL;

	K_node* child = Root;
	K_node* temp, *temp1=NULL;
	int i=0;

	while(i<len)
	{
		for(temp=child->leftchild; temp!=NULL; temp = temp->rightsibling)
		{
			if(temp->itemName==keySet[i])break;
			if(temp->rightsibling==NULL)temp1 = temp;
		}

		if(!temp)break;

		child=temp;
		i++;
	}
	
	if(child != Root && child->leftchild == NULL)
		return NULL; //target keyset is superset of one of the exisitn key

	while(i<len)
	{   
		child = child->append(this, temp1, keySet[i]);
		i++;
	}
	child->next = header;
	header = child;

	return child;
}

int K_tree::outputKeys(FSout* fsout, int maxlen)
{
	if(fsout == NULL)
		return 0;

	K_node *temp1, *temp;
	temp = header;
	int *key = new int[maxlen];
	int keylen;

	while(temp !=NULL)
	{
		keylen =0;
		temp1 = temp;
		while(temp1 != Root)
		{
			key[keylen++] = temp1->itemName;
			//fsout->printItem(temp1->itemName);
			temp1 = temp1->par;
		}
		for(;keylen>0;keylen--)
			fsout->printItem(key[keylen-1]);

		fsout->printString(";");
		temp = temp->next;
	}

	delete []key;

	return 1;
}

int K_tree::mergeTree(K_tree *new_tree, int maxlen)
{
	if(new_tree==NULL)
		return 0;
	
	K_node* temp=new_tree->header;
	K_node *temp1;
	int *iset = new int[maxlen];
	int i;
	
	while(temp!=NULL)
	{
		i=0;
		temp1 = temp;
		
		while(temp1 != new_tree->Root)
		{
			iset[i++] = temp1->itemName;
			temp1 = temp1->par;	
		}
		insertKey(iset, i);
		temp = temp->next;
	}

	return 1;
}

//////////////////////////////////////////////////////////////////////
// Functions of T_tree Class
//////////////////////////////////////////////////////////////////////

void T_tree::init()
{
	Root = (Tnode*)fp_buf1->newbuf(1, sizeof(Tnode));
	Root->init(NULL, -1, 0);

	header = NULL;
}

T_tree::~T_tree()
{
}

/*void T_tree::scan_DB(Data *fdat)
{
	int i;
	Transaction *Tran = new Transaction;
	assert(Tran!=NULL);
	
	for(i=0; i<3; i++)
	{
		fdat->getNextTransaction(Tran);
		switch(i)
		{
			case 0:
			TRANSACTION_NO = Tran->t[0];
			break;
		case 1:
			ITEM_NO = Tran->t[0];
			break;
		case 2:
			order_item = (int*) fp_buf1->newbuf(ITEM_NO, sizeof(int));
			memcpy(order_item, Tran->t, sizeof(int)*ITEM_NO);
			break;
		}
	}
	
	patternTable=new hash_table(ITEM_NO, ITEM_NO);
	header = (Tnode**)fp_buf1->newbuf(TRANSACTION_NO, sizeof(Tnode*));
	for(i=0; i<TRANSACTION_NO;i++)
	{
		//header[i] = (Tnode*)fp_buf1->newbuf(1, sizeof(Tnode));
		//header[i]->init(NULL, i);
		header[i] = NULL;
	}

	int* tidset, * closeSet;
	int tidlen,index;
	int support=0;
	Tnode* child;
	K_tree* ktree = NULL;

	while(Tran = fdat->getNextTransactionWTid(Tran, tidset, tidlen, ktree))
	{	
		if(tidset && tidlen >0)
		{
			closeSet =(int*)fp_buf1->newbuf(Tran->length, sizeof(int));
			memcpy(closeSet, Tran->t, sizeof(int)*Tran->length);
			
			index = patternTable->insertEC(closeSet, Tran->length, tidlen, ktree);
			child = insertTidSet(tidset, tidlen, index);
			patternTable->insertTidPointer(index, child);

			ktree = NULL;
			
			delete []tidset;
		}
		else
		{
			fullItem =(int*)fp_buf1->newbuf(Tran->length, sizeof(int));
			fullNum = Tran->length;
			memcpy(fullItem, Tran->t, sizeof(int)*Tran->length);
		}
		
	}
	delete Tran;

}*/

int T_tree::scan_FP(FK_tree *fptree)
{
	if(fptree == NULL)
		return 0;
	
	int closeSize = patternTable->getCloseSize();
	LPCloseHashNode* closeTable = patternTable->getCloseTable();
	
	header = (Tnode**)fp_buf1->newbuf(TRANSACTION_NO, sizeof(Tnode*));
	for(int i=0; i<TRANSACTION_NO;i++)
	{
		header[i] = NULL;
	}
	
	for(i=0; i<closeSize;i++)
	{
		LPCloseHashNode lpCloseNode=closeTable[i];
		if(lpCloseNode==NULL || lpCloseNode->support == TRANSACTION_NO)
			continue;

		int *tidset = new int[lpCloseNode->support];
		int tidlen = 0;
		fptree->scanTid(lpCloseNode->iset, lpCloseNode->length, tidset, lpCloseNode->support, tidlen);

		Tnode* temp = NULL;
		temp = insertTidSet(tidset, tidlen, i);
		if(temp == NULL)
		{
			printf("tid-tree building error!");
			scanf("exit");
			exit(1);
		}
		patternTable->insertTidPointer(i, temp);
		delete []tidset;
	}

	return 1;
}

Tnode* T_tree::insertTidSet(int *tidSet, int tidlen, int ECindex)
{
	if(tidSet == NULL)
		return NULL;

	Tnode* child = Root;
	Tnode* temp, *temp1=NULL;
	int i=0;

	while(i<tidlen)
	{
		for(temp=child->leftchild; temp!=NULL; temp = temp->rightsibling)
		{
			if(temp->tidName==tidSet[i])break;
			if(temp->rightsibling==NULL)temp1 = temp;
		}

		if(!temp)break;

		child=temp;
		i++;
	}

	while(i<tidlen)
	{   
		child = child->append(this, temp1, tidSet[i], -255);
		i++;
	}

	child->ECindex = ECindex;
	return child;
}

int T_tree::removeTrans(int *tidSet, int tidlen)
{
	if(tidSet == NULL)
		return 0;
	
	int i;
	Tnode *current, *par, *rs, *lc, *temp, *temp1;

	for(i=0; i<tidlen; i++)
	{
		current = header[tidSet[i]];
		
		while(current!=NULL)
		{
			par = current->par;
			rs =  current->rightsibling;
			lc = current->leftchild;
			
			//merge ECs
			if(current->ECindex>=0)
			{
				//par: par got ec ==> merge ecs
				if(par->ECindex>=0)
				{
					par->ECindex = patternTable->mergeECs(current->ECindex, par->ECindex);
				}
				else //case: par no ec index
					par->ECindex = current->ECindex;
				
			}

			//remove current node
			//case: when current is the first child
			if(par->leftchild == current)
			{
				if(lc == NULL)
					par->leftchild = rs;
				else
				{
					par->leftchild = lc;
					lc->par = par;

					temp = lc;
					while(temp->rightsibling!=NULL)
					{
						temp->par = par;
						temp = temp->rightsibling;
					}

					temp->rightsibling = rs;
				}
			}
			else //case: when current is not the first child
			{
				temp1 = par->leftchild;

				while(temp1!=NULL && temp1->rightsibling!=current)
					temp1 = temp1->rightsibling;

				if(lc==NULL)
					temp1->rightsibling = rs;
				else
				{
					temp1->rightsibling=lc;
					lc->par = par;

					temp = lc;
					while(temp->rightsibling!=NULL)
					{
						temp->par = par;
						temp = temp->rightsibling;
					}

					temp->rightsibling = rs;
				}	
			}

			current = current->next;
		}
	}

	return 1;
}

int T_tree::outputTidSet(int *&tidset, Tnode *tidNode, int old_support)
{
	if(tidset==NULL || tidNode == NULL)
		return 0;
	
	Tnode* temp = tidNode;
	int support =0, i=old_support;
	//int *tidset = new int[old_support];

	while(temp!=Root)
	{
		tidset[--i] = temp->tidName; //output tid list is in acending order
		temp = temp->par;
	}
	
	support = old_support-i;

	//delete []tidset;
	return support;
}

int T_tree::outputTree(FSout *fsout)
{
	if(fsout == NULL)
		return 0;
	
	fsout->printString("$ "); //$ stands for root

	Tnode* temp = Root;
	
	printTree(fsout, temp->leftchild);

	return 1;
}

int T_tree::printTree(FSout *fsout, Tnode* current)
{
	if(fsout == NULL || current == NULL)
		return 0;
	
	Tnode* temp = current;
	while(temp != NULL)
	{
		fsout->printDigit(temp->tidName);
		fsout->printString(" ");
		
		if(temp->ECindex>=0)
		{
			fsout->printString("[");
			fsout->printDigit(temp->ECindex);
			fsout->printString("] ");
		}	

		if(current->leftchild)
			printTree(fsout, temp->leftchild);

		fsout->printString("/");
		temp = temp->rightsibling;
	}
	
	//fsout->printString("/");
	return 1;
}