/*----------------------------------
File     : data.cpp
Contents : data set management
----------------------------------*/

#include "data.h"
#include "common.h"
#include "tree.h"

void Transaction::DoubleTrans(int item)
{
	int* temp = new int[2*item];
	maxlength = 2*item;
	for(int i=0; i<length; i++)
		temp[i] = t[i];
	delete []t;
	t = temp;
}

Data::Data(char *filename)
{
#ifndef BINARY
	in = fopen(filename,"rt");
#else
	in = fopen(filename, "rb");
#endif
}

Data::~Data()
{
	if(in) fclose(in);
}

int Data::isOpen()
{
	if(in) return 1;
	else return 0;
}

Transaction *Data::getNextTransaction(Transaction* Trans)
{
	Trans->length = 0;
	
	// read list of items
#ifndef BINARY	  
	char c;
	int item, pos;
	do {
		item=0;
		pos=0;
		c = getc(in);
		while((c >= '0') && (c <= '9')) {
			item *=10;
			item += int(c)-int('0');
			c = getc(in);
			pos++;
		}
		if(pos)
		{
			if(Trans->length >= Trans->maxlength)
				Trans->DoubleTrans(Trans->length);
			
			Trans->t[Trans->length] = item;
			Trans->length++;
		}
	}while(c != '\n' && !feof(in));
	// if end of file is reached, rewind to beginning for next pass
	if(feof(in)){
		rewind(in);
		
		//debug for memory leak 21/10/2004
		delete Trans; 
		
		return 0;
	}
	// Note, also last transaction must end with newline, 
	// else, it will be ignored
#else
	int i, nitem, *buffer=new int;
	fread((char*)buffer, sizeof(int), 1, in);
	if(feof(in))
	{
		rewind(in);
		
		//debug for memory leak 21/10/2004
		delete Trans; 
		
		return 0;
	}
	fread((char*)buffer, sizeof(int), 1, in);
	fread((char*)buffer, sizeof(int), 1, in);
	nitem=*buffer;
	for(i=0; i<nitem; i++)
	{
		fread((char*)buffer, sizeof(int), 1, in);
		
		if(Trans->length >= Trans->maxlength)
			Trans->DoubleTrans(Trans->length);
		
		Trans->t[Trans->length] = *buffer;
		Trans->length++;
	}
#endif
	
	return Trans;
}

Transaction *Data::getNextTransactionWTid(Transaction* Trans, int &support, int& index, K_tree *& ktree)
{
	Trans->length = 0;
	
	char c;
	int item, pos;
	
	//capture index
	do{
		c = getc(in);
	}while(c!='[' && c != '\n' && !feof(in));

	do{
		item=0;
		pos=0;
		c = getc(in);
		while((c >= '0') && (c <= '9')) {
			item *=10;
			item += int(c)-int('0');
			c = getc(in);
			pos++;
		}
		if(pos)
			index = item;

	}while(c !=']' && c !='\n' && !feof(in));

	
	//capture closed patterns
	do{
		item=0;
		pos=0;
		c = getc(in);
		while((c >= '0') && (c <= '9')) {
			item *=10;
			item += int(c)-int('0');
			c = getc(in);
			pos++;
		}
		if(pos)
		{
			if(Trans->length >= Trans->maxlength)
				Trans->DoubleTrans(Trans->length);
			
			Trans->t[Trans->length] = item;
			Trans->length++;
		}
	}while(c !='(' && c !='\n' && !feof(in));
	
	//capture support value
	do{
		item=0;
		pos=0;
		c = getc(in);
		while((c >= '0') && (c <= '9')) {
			item *=10;
			item += int(c)-int('0');
			c = getc(in);
			pos++;
		}
		if(pos)
			support = item;
	}while(c !=')' && c !='\n' && !feof(in));
	
	//capture key patterns
	int* key = new int[Trans->length];
	int keylen;
	
	do{
		keylen = 0;
		do{
			item=0;
			pos=0;
			c = getc(in);
			while((c >= '0') && (c <= '9')) {
				item *=10;
				item += int(c)-int('0');
				c = getc(in);
				pos++;
			}
			if(pos)
				key[keylen++] = item;

		}while(c !=';' && c !='\n' && !feof(in));
		
		if(keylen)
		{
			if(ktree==NULL)
			{
				ktree = (K_tree*)fp_buf->newbuf(1, sizeof(K_tree));
				ktree->init();
			}

			ktree->insertKey(key, keylen);
		}

	}while(c !='\n' && !feof(in));

	delete []key;
	
	// if end of file is reached, rewind to beginning for next pass
	if(feof(in)){
		rewind(in);
		
		//debug for memory leak 21/10/2004
		delete Trans; 
		
		return 0;
	}
	
	return Trans;
}

int Data::getTtree(T_tree *ttree)
{
	if(ttree==NULL)
		return 0;

	char c;
	int item, pos;
	Tnode* current = ttree->Root;
	Tnode* sib = NULL;
	
	c = getc(in);
	
	do{	
		
		do{
			item=0;
			pos=0;
			while((c >= '0') && (c <= '9')) {
				item *=10;
				item += int(c)-int('0');
				c = getc(in);
				pos++;
			}
			//insert node
			if(pos)
			{
				if(current->leftchild)
				{
					for(sib=current->leftchild; sib->rightsibling!=NULL; sib = sib->rightsibling);
				}
				current = current->append(ttree, sib, item);
			}

			c = getc(in);

		}while(c!='/' &&c!='[' && !feof(in));

		if(c=='[')
		{
			do{
				item=0;
				pos=0;
				c = getc(in);
				while((c >= '0') && (c <= '9')) {
					item *=10;
					item += int(c)-int('0');
					c = getc(in);
					pos++;
				}
				//insert index
				if(pos)
				{
					current->ECindex = item;
					patternTable->insertTidPointer(item, current);
				}
				
			}while(c!=']');
		}

		while(c == '/')
		{
			current = current->par;
			c = getc(in);
		}

	}while(!feof(in));

	return 1;
}