// hash_table.cpp: implementation of the hash_table class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include "hash_table.h"
#include "math.h"
#include "common.h"
#include "tree.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
#define MAXTRY 5
#define INCREMENT(x) x+x+x
#define MAXTIMES 512

extern bool isSubset(int* controlSet, int conlen, int* caseSet, int caselen);

hash_table::hash_table()
{
   closeSize=0;
   closeThreshold=0;
   currCloseNum=0;
   mapSize=0;
   closeMap=NULL;

   //secModMap=NULL;
   closeTable=NULL;

   int initSize=INITSIZE;

   //int estSize=TRANSACTION_NO*itemNum*1024;
   //if(estSize<initSize)
   //initSize=getPrime(INITSIZE);

   //cout<<"hash table size is "<<initSize<<endl;
   printf("hash table size is %d\n",initSize);
   //int initSize=getPrime(TRANSACTION_NO*2);
   
   //initCloseTable(initSize);
   //initModMap(totalNum,closeMap,initSize);

   //for debug
   numInsert=0;
   numProbe=0;
 }

void hash_table::init(int totalNum)
{
	int initSize=INITSIZE;
	initCloseTable(initSize);
	initModMap(totalNum,closeMap,initSize);
}

hash_table::~hash_table()
{
   releaseTable();
   if(closeMap!=NULL)
	   delete closeMap;
   if(numInsert!=0)
	   //cout<<"insert "<<numInsert<<" probe "<<numProbe<<" average "<<numProbe*1.0/numInsert<<endl;
		printf("insert %d probe %d average %f\n",numInsert,numProbe, numProbe*1.0/numInsert);
}

bool hash_table::initCloseTable(int newSize)
{
  closeTable=new LPCloseHashNode[newSize];
  for( int i=0;i<newSize;i++)
  {
     closeTable[i]=NULL; 
  }
  closeSize=newSize;
  currCloseNum=0;

  closeThreshold=LOADFACTOR*closeSize;

  return true;
}

bool hash_table::releaseTable()
{
  for(int i=0;i<closeSize;i++)
  {
    if(closeTable[i]!=NULL)
    {
	   //remove keys
		
		//remove closed

	   //delete closeTable[i]->iset;   
	   delete closeTable[i];
	   currCloseNum--;
    }
  }
  delete closeTable;
  closeTable=NULL;
  closeSize=0;
  return true;
}

int  hash_table::single_mod(int i, int prime)
{
  int ret=-1;
  if(i<0) return -1;
  unsigned long pValue=0;
  if(i<MAXPOWER)
    ret=(unsigned long)(pow(2,i)) % prime;
  else
    ret=(2*single_mod(i-1, prime)) % prime;

  if(ret<0)
  {
	  //cout<<"single mod return negative"<<ret<<endl;
	  printf("single mod return negative %d", ret);
  }
  return ret;
}

int  hash_table::set_mod(int *iset, int length, int order, int *map, int size)
{
  int sum=0;
  
  for(int i=0;i<length;i++)
  {
	//if(order==0)
	if(iset[i]>=0) //ignore removed transactions
	  sum=sum+map[iset[i]];
    while(sum>=size)
      sum=sum-size;
  	//else
    //     sum=sum+secModMap[iset[i]];
  }
  //if(order==0)
  //else
  //  ret=sum % (size+OFFSET);
  
  return sum;
}

bool hash_table::compareKey(int *iset1, int length1, int *iset2, int length2)
{
  if(length1!=length2)
	  return false;
  for(int i=length1-1;i>=0;i--)
  {
	  if(iset1[i]!=iset2[i]) return false;
  }
  return true;
}

int   hash_table::insert(int * & iCloseSet, int cLen, int support, int posCount, int *iKeySet, int kLen)
{
  int i=0;
  int initIndex=0;
  int index;
  initIndex=set_mod(iCloseSet,cLen,0,closeMap, closeSize);
  index=initIndex;
  
  KeySetNode * lpKeyNode=new KeySetNode;
  lpKeyNode->iset=iKeySet;
  lpKeyNode->length=kLen;
  lpKeyNode->next=NULL;

  while(i<closeSize)
  {
	if(closeTable[index]==NULL)
    {
	    LPCloseHashNode lpNode=new CloseHashNode;
		lpNode->iset=iCloseSet;
		lpNode->length=cLen;
		lpNode->support=support;
		lpNode->posCount=posCount;
		//lpNode->head=lpKeyNode;
		closeTable[index]=lpNode;
		currCloseNum++;
		if(currCloseNum>closeThreshold)
        {
			//for debug
			//cout<<"now rehash close hash table from "<<currCloseNum<<endl;
			printf("now rehash close hash table from %d\n", currCloseNum);
			
			rehashCloseTable();
        }
        return index;
	}else
    {
        LPCloseHashNode lpNode=closeTable[index];
		bool bEqual=compareKey(iCloseSet,cLen, lpNode->iset, lpNode->length);
		if(bEqual)
        {
           //lpKeyNode->next=lpNode->head;
		   //lpNode->head=lpKeyNode;
           delete iCloseSet; //added by mengling for memory leak
		   return index;
        }    
	}
    i++;
  	index=initIndex+INCREMENT(i);  
    while(index>=closeSize)
	{
      index=index-closeSize;
	}

  }

  return -1;
}

int  hash_table::insert(LPCloseHashNode lpCloseNode)
{
  int i=0;
  int initIndex=0;
  int index;
  
  initIndex=set_mod(lpCloseNode->iset,lpCloseNode->length,0, closeMap,closeSize);
  index=initIndex;
  
  while(i<closeSize)
  {
	if(closeTable[index]==NULL)
    {
		currCloseNum++;
		closeTable[index]=lpCloseNode;
		if(currCloseNum>closeThreshold)
        {
			//for debug
			//cout<<"now rehash close table from "<<currCloseNum<<endl;
			printf("now rehash close table from %d\n",currCloseNum);
			rehashCloseTable();
        }
        return index;
	}else
    {
        LPCloseHashNode lpCurrNode=closeTable[index];
		bool bEqual=compareKey(lpCloseNode->iset,lpCloseNode->length,lpCurrNode->iset, lpCurrNode->length);
		if(bEqual)
        {
           /*KeySetNode *lpKeyNode=lpCurrNode->head;
		   while(lpKeyNode->next!=NULL)
              lpKeyNode=lpKeyNode->next;
           lpKeyNode->next=lpCloseNode->head;*/
           return index;
        }    
    }
    i++;
  	index=initIndex+INCREMENT(i);  
    while(index>=closeSize)
	{
      index=index-closeSize;
	}

  }

  return -1;
}

bool hash_table::rehashCloseTable()
{
  LPCloseHashNode *oldTable=closeTable;
  int oldSize=closeSize;
  closeSize=getPrime(2*closeSize);
  initCloseTable(closeSize);
  initModMap(mapSize,closeMap,closeSize);
  for(int i=0;i<oldSize;i++)
  {
    if(oldTable[i]==NULL)
		continue;
    insert(oldTable[i]);  
  }
  delete oldTable;
  return true;
}
 
int  hash_table::getPrime(int start)
{
  int iCurr=start;
  while(1)
  {
    if(isPrime(iCurr))
		return iCurr;
	iCurr++;
  }
  
  return -1;
}

bool hash_table::isPrime(int num)
{
  if(num<=2) return false;

  for(int i=2;i<(num/2);i++)
	  if((num %i)==0) return false;

  return true;
}


bool  hash_table::initModMap(int num, int * & map, int prime)
{
  if(mapSize>0)
  {
    if(map!=NULL)
		delete map;
	//if(secModMap!=NULL)
	//	delete secModMap;
  }
  map=new int[num];
  //secModMap=new int[num];
  mapSize=num;
  for(int i=0;i<num;i++)
	  map[i]=single_mod(i,prime);

  return true;
}

bool  hash_table::begin()
{
  curr=0;

  return true;
}


bool  hash_table::outputClose(FSout *fout, double & keyNum, double &closeNum, T_tree* ttree)
{
  for(int i=0; i<closeSize;i++)
  {
     LPCloseHashNode lpCloseNode=closeTable[i];
	 if(lpCloseNode==NULL)
		 continue;
	 
	 //output support
	 int *tidset = new int[lpCloseNode->support];
	 int support = 0;
	 support = ttree->outputTidSet(tidset, lpCloseNode->tids, lpCloseNode->support);
	
	 if(support < THRESHOLD)
	 {
		 delete []tidset;
		 continue;
	 }

	 //output closed pattern
     fout->printOrderSet(lpCloseNode->length, lpCloseNode->iset, 0);

	 fout->printString("(");
	 fout->printDigit(support);
	 fout->printString(")");

	 //output key patterns
	 K_tree* ktree = lpCloseNode->keys;
	 if(ktree)
		 ktree->outputKeys(fout, lpCloseNode->length);

	 fout->printString("|");
	
	 //output tidlist
	 fout->printTid(&tidset[lpCloseNode->support-support], support);

	 delete []tidset;
	 //fout->printTid(lpCloseNode->TidSet, lpCloseNode->TidLength);
	 //fout->printTime("Tid:",lpCloseNode->TidLength);
	 //fout->printString("\n");
	 closeNum++;
  }
  return true;
}

bool  hash_table::outputClose(FSout *fout, double & keyNum, double &closeNum)
{
	for(int i=0; i<closeSize;i++)
	{
		LPCloseHashNode lpCloseNode=closeTable[i];
		if(lpCloseNode==NULL)
			continue;
		
		if(lpCloseNode->support < THRESHOLD)
			continue;
		//output index
		fout->printString("[");
		fout->printDigit(i);
		fout->printString("]");

		//output closed pattern
		fout->printOrderSet(lpCloseNode->length, lpCloseNode->iset, 0);
		
		if(lpCloseNode->tids)
			lpCloseNode->support = lpCloseNode->tids->depth;
		else
			lpCloseNode->support = TRANSACTION_NO;
		
		fout->printString("(");
		fout->printDigit(lpCloseNode->support);
		fout->printString(")");
		
		//output key patterns
		K_tree* ktree = lpCloseNode->keys;
		if(ktree)
			ktree->outputKeys(fout, lpCloseNode->length);

		fout->printString("\n");
		closeNum++;
	}
	return true;
}

int hash_table::insertEC(int *&iCloseSet, int clen, int support, K_tree *keys)
{
	if(iCloseSet == NULL)
		return 0;
	
	int i=0;
	int initIndex=0;
	int index;
	initIndex=set_mod(iCloseSet,clen,0,closeMap, closeSize);
	index=initIndex;
	
	while(i<closeSize)
	{
		if(closeTable[index]==NULL)
		{
			LPCloseHashNode lpNode=new CloseHashNode;
			lpNode->iset=iCloseSet;
			lpNode->length=clen;
			lpNode->support=support;
			lpNode->keys = keys;
			lpNode->tids = NULL;
			closeTable[index]=lpNode;
			currCloseNum++;
			if(currCloseNum>closeThreshold)
			{
				//for debug
				printf("now rehash close hash table from %d\n", currCloseNum);
				
				rehashCloseTable();
			}
			return index;
		}

		i++;
		index=initIndex+INCREMENT(i);  
		while(index>=closeSize)
		{
			index=index-closeSize;
		}
		
	}

	return 0;
}

int hash_table::insertTidPointer(int index, Tnode* tidPt)
{
	if(tidPt == NULL || closeTable[index]==NULL)
		return 0;
	
	LPCloseHashNode lpNode = closeTable[index];
	lpNode->tids = tidPt;

	return index;
}

int hash_table::mergeECs(int index1, int index2)
{
	LPCloseHashNode lpCloseNode1=closeTable[index1];
	LPCloseHashNode lpCloseNode2=closeTable[index2];

	if(lpCloseNode1 == NULL || lpCloseNode2 == NULL)
		return -255;
	
	if(lpCloseNode1->length>lpCloseNode2->length)
	{
		bool sub = isSubset(lpCloseNode1->iset, lpCloseNode1->length,
			lpCloseNode2->iset, lpCloseNode2->length);

		//remove infreq ecs
		if(sub==false)
		{
			lpCloseNode1->tids->ECindex = -255;
			lpCloseNode2->tids->ECindex = -255;

			deleteEC(index1);
			deleteEC(index2);

			return -255;
		}
		
		//merge keys
		K_tree *temp = lpCloseNode1->keys;
		lpCloseNode1->keys = lpCloseNode2->keys;
		lpCloseNode1->keys->mergeTree(temp, lpCloseNode1->length);

		delete lpCloseNode2;
		closeTable[index2] = NULL;

		return index1;
	}
	else
	{
		bool sub = isSubset(lpCloseNode2->iset, lpCloseNode2->length,
			lpCloseNode1->iset, lpCloseNode1->length);

		//remove infreq ecs
		if(sub==false)
		{
			lpCloseNode1->tids->ECindex = -255;
			lpCloseNode2->tids->ECindex = -255;

			deleteEC(index1);
			deleteEC(index2);

			return -255;
		}
		
		//merge keys
		K_tree *temp = lpCloseNode2->keys;
		lpCloseNode2->keys = lpCloseNode2->keys;
		lpCloseNode2->keys->mergeTree(temp, lpCloseNode2->length);

		delete lpCloseNode1;
		closeTable[index1] = NULL;

		return index2;
	}
}

int hash_table::readClose(Data *fdat)
{
	if(fdat == NULL)
	{
		printf("input file error!");
		scanf("exit");
		exit(1);
	}

	int i;
	Transaction *Tran = new Transaction;
	//assert(Tran!=NULL);
	
	//capture dataset informations
	for(i=0; i<3; i++)
	{
		fdat->getNextTransaction(Tran);
		switch(i)
		{
			case 0:
			TRANSACTION_NO = Tran->t[0];
			break;
		case 1:
			ITEM_NO = Tran->t[0];
			break;
		case 2:
			order_item = (int*) fp_buf->newbuf(ITEM_NO, sizeof(int));
			memcpy(order_item, Tran->t, sizeof(int)*ITEM_NO);
			break;
		}
	}
	//initialize hash_table
	init(ITEM_NO);

	int* closeSet;
	int support=0, index = -255;
	K_tree* ktree = NULL;

	while(Tran = fdat->getNextTransactionWTid(Tran, support, index, ktree))
	{	
		closeSet =(int*)fp_buf->newbuf(Tran->length, sizeof(int));
		memcpy(closeSet, Tran->t, sizeof(int)*Tran->length);
			
		LPCloseHashNode lpNode=new CloseHashNode;
		lpNode->iset = closeSet;
		lpNode->length = Tran->length;
		lpNode->support = support;
		lpNode->keys = ktree;
		lpNode->tids = NULL;

		closeTable[index] = lpNode;
		
		ktree = NULL;
		
	}
	delete Tran;

	return 1;
}

void hash_table::deleteEC(int index)
{
	LPCloseHashNode lpCloseNode = closeTable[index];
	
	delete lpCloseNode;
	closeTable[index] = NULL;
}