#include <assert.h>
#include "tree.h"
#include "hash_table.h"
#include "buffer.h"

//////////////////////////////////////////////////////////////////////
// external variables
//////////////////////////////////////////////////////////////////////

extern memory* fp_buf;
extern int TRANSACTION_NO;
extern int THRESHOLD;
extern int ITEM_NO;
extern int* order_item;
extern hash_table *patternTable;
extern int* fullItem;
extern int fullNum;


//////////////////////////////////////////////////////////////////////
// Functions of K_tree Class
//////////////////////////////////////////////////////////////////////


void K_tree::init()
{
	Root = (Knode*)fp_buf->newbuf(1, sizeof(Knode));
	Root->init(NULL, -1);

	header = NULL;
}

K_tree::~K_tree()
{
}

Knode* K_tree::insertKey(int *keySet, int len)
{
	if(keySet==NULL)
		return NULL;

	Knode* child = Root;
	Knode* temp, *temp1=NULL;
	int i=0;

	while(i<len)
	{
		for(temp=child->leftchild; temp!=NULL; temp = temp->rightsibling)
		{
			if(temp->itemName==keySet[i])break;
			if(temp->rightsibling==NULL)temp1 = temp;
		}

		if(!temp)break;

		child=temp;
		i++;
	}
	
	if(child != Root && child->leftchild == NULL)
		return NULL; //target keyset is superset of one of the exisitn key
	
	//got bug
	/*if(i>=len && child && child->leftchild)
	{
		bool debug = true;
	}*/

	while(i<len)
	{   
		child = child->append(this, temp1, keySet[i]);
		i++;
	}
	child->next = header;
	header = child;

	return child;
}

int K_tree::outputKeys(FSout* fsout, int maxlen)
{
	if(fsout == NULL)
		return 0;

	Knode *temp1, *temp;
	temp = header;
	int *key = new int[maxlen];
	int keylen;

	while(temp !=NULL)
	{
		keylen =0;
		temp1 = temp;
		while(temp1 != Root)
		{
			key[keylen++] = temp1->itemName;
			//fsout->printItem(temp1->itemName);
			temp1 = temp1->par;
		}
		for(;keylen>0;keylen--)
			fsout->printItem(key[keylen-1]);

		fsout->printString(";");
		temp = temp->next;
	}

	delete []key;

	return 1;
}

int K_tree::mergeTree(K_tree *new_tree, int maxlen)
{
	if(new_tree==NULL)
		return 0;
	
	Knode* temp=new_tree->header;
	Knode *temp1;
	int *iset = new int[maxlen];
	int i;
	
	while(temp!=NULL)
	{
		i=0;
		temp1 = temp;
		
		while(temp1 != new_tree->Root)
		{
			iset[i++] = temp1->itemName;
			temp1 = temp1->par;	
		}
		insertKey(iset, i);
		//printf("insert key\n");
		temp = temp->next;
	}

	//printf("merge finished\n");

	return 1;
}

//////////////////////////////////////////////////////////////////////
// Functions of T_tree Class
//////////////////////////////////////////////////////////////////////

void T_tree::init()
{
	Root = (Tnode*)fp_buf->newbuf(1, sizeof(Tnode));
	Root->init(NULL, -1);

	header = NULL;
}

T_tree::~T_tree()
{
}

void T_tree::scan_DB(Data *fdat)
{
	header = (Tnode**)fp_buf->newbuf(TRANSACTION_NO, sizeof(Tnode*));
	for(int i=0; i<TRANSACTION_NO;i++)
		header[i] = NULL;
	
	fdat->getTtree(this);
}

Tnode* T_tree::insertTidSet(int *tidSet, int tidlen, int ECindex)
{
	if(tidSet == NULL)
		return NULL;

	Tnode* child = Root;
	Tnode* temp, *temp1=NULL;
	int i=0;

	while(i<tidlen)
	{
		for(temp=child->leftchild; temp!=NULL; temp = temp->rightsibling)
		{
			if(temp->tidName==tidSet[i])break;
			if(temp->rightsibling==NULL)temp1 = temp;
		}

		if(!temp)break;

		child=temp;
		i++;
	}

	while(i<tidlen)
	{   
		child = child->append(this, temp1, tidSet[i], -255);
		i++;
	}

	child->ECindex = ECindex;
	return child;
}

int T_tree::removeTrans(int *tidSet, int tidlen)
{
	if(tidSet == NULL)
		return 0;
	
	int i;
	Tnode *current, *par, *rs, *lc;

	for(i=0; i<tidlen; i++)
	{
		//printf("%d %d\n", i, tidSet[i]);
		
		current = header[tidSet[i]];
		
		while(current!=NULL)
		{
			par = current->par;
			rs =  current->rightsibling;
			lc = current->leftchild;
			
			//merge ECs
			mergeECs(par, current);

			//remove current node
			removeNode(current);

			current = current->next;
		}
	}

	//updateDepth(Root);

	return 1;
}

int T_tree::mergeECs(Tnode* dst, Tnode* src)
{
	if(dst == NULL || src == NULL)
		return 0;
	
	if(src->ECindex>=0)
	{
		if(dst->depth<THRESHOLD)
		{
			patternTable->deleteEC(src->ECindex);
			return 1;
		}
		
		//dst: dst got ec ==> merge ecs
		if(dst->ECindex>=0)
		{
			dst->ECindex = patternTable->mergeECs(src->ECindex, dst->ECindex);
		}
		else //case: dst no ec index
			dst->ECindex = src->ECindex;
		
	}
	
	return 1;
}

int T_tree::removeNode(Tnode* current)
{
	if(current == NULL)
		return 0;
	Tnode* par =  current->par;

	//case: current is par's first child
	if(par->leftchild == current)
	{
		par->leftchild = current->rightsibling;
	}
	else //case: when current is not the first child
	{
		Tnode* temp1 = par->leftchild;
		
		while(temp1!=NULL && temp1->rightsibling!=current)
		{
			/*if(temp1->par != current->par)
			{
				bool debug = true;
			}*/
			temp1 = temp1->rightsibling;
		}
		
		
		temp1->rightsibling = current->rightsibling;
	}
	
	//remove current from header table
	if(current->pre)
	{
		current->pre->next = current->next;
		if(current->next)
			current->next->pre = current->pre;
	}
	else
	{
		header[current->tidName] = current->next;
		if(current->next)
			current->next->pre = NULL;
	}

	if(current->leftchild)
	{
		Tnode* temp = current->leftchild;
		Tnode* temp1, *temp2;

		while(temp != NULL)
		{
			temp1 = temp->rightsibling;

			temp2 = par->appendPath(this, temp);
			temp = temp1;
		}
	}

	return 1;
}

int T_tree::updateDepth(Tnode* current)
{
	if(current == NULL)
		return 0;

	Tnode* temp = current->leftchild;

	while(temp)
	{
		temp->depth = current->depth+1;

		updateDepth(temp);

		temp = temp->rightsibling;
	}

	return 1;
}

int T_tree::outputTidSet(int *&tidset, Tnode *tidNode, int old_support)
{
	if(tidset==NULL || tidNode == NULL)
		return 0;
	
	Tnode* temp = tidNode;
	int support =0, i=old_support;
	//int *tidset = new int[old_support];

	while(temp!=Root)
	{
		tidset[--i] = temp->tidName; //output tid list is in acending order
		temp = temp->par;
	}
	
	support = old_support-i;

	//delete []tidset;
	return support;
}
int T_tree::outputTree(FSout *fsout)
{
	if(fsout == NULL)
		return 0;
	
	fsout->printString("$ "); //$ stands for root

	Tnode* temp = Root;
	
	printTree(fsout, temp->leftchild);

	return 1;
}

int T_tree::printTree(FSout *fsout, Tnode* current)
{
	if(fsout == NULL || current == NULL)
		return 0;
	
	Tnode* temp = current;
	while(temp != NULL)
	{
		
		fsout->printDigit(temp->tidName);
		fsout->printString(" ");
		
		if(temp->ECindex>=0)
		{
			fsout->printString("[");
			fsout->printDigit(temp->ECindex);
			fsout->printString("] ");
		}	
		
		if(temp->leftchild)
			printTree(fsout, temp->leftchild);

		fsout->printString("/");
		temp = temp->rightsibling;
	}
	
	//fsout->printString("/");
	return 1;
}