#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <time.h>
#include "common.h"
#include "buffer.h"
#include "hash_table.h"

#include "tree.h"

#define LINT sizeof(int)

int* order_item;   // given order i, order_item[i] gives the original itemname, corresponding to table
int* item_order;   // given item i, item_order[i] gives its new order, corresponding to order
				   // order_item[item_order[i]]=i; item_order[order_item[i]]=i;

int* fullItem=NULL;     // the items with full support, maybe item, maybe the index.
int  fullNum=0;

hash_table *patternTable;

int TRANSACTION_NO=0;
int POSITIVE_NO=0;
int ITEM_NO=0;
int MAX_ITEM=0;
int THRESHOLD;
double RATIO_THRESHOLD=0;
int TIDLENGTH = 0;
int TOTAL_UPDATE=0;

memory* fp_buf;
//memory* hash_buf;

bool isSubset(int* controlSet, int conlen, int* caseSet, int caselen)
{
	if(controlSet == NULL || caseSet == NULL
		|| conlen < caselen)
		return false;

	int *map = new int[ITEM_NO];
	int i;

	for(i=0;i<ITEM_NO; i++)
		map[i] = -1;

	for(i=0; i< conlen; i++)
	{
		if(controlSet[i]<ITEM_NO)
			map[controlSet[i]] = 1;
	}

	for(i=0;i<caselen; i++)
	{
		if(caseSet[i] <ITEM_NO && map[caseSet[i]]==-1)
		{
			delete []map;
			return false;
		}
	}
	delete []map;
	return true;
}

void main(int argc, char **argv)
{
	//bool   bClass=false;
	//int    posItem=-1;
	char *inFileName;
	char *outFileName;
	double threshold;
	//RATIO_THRESHOLD = 0;
	int    nFull=0;
	int i=1, index =0;
	int nofremovedT;
	int startPos;

		if (argc < 6)
	{
		cout << "usage: trum <infile> <MINSUP> <outfile> <nofremoved> <startPos>\n";
		exit(1);
	}
    while(i<argc)
    {

		switch(index++)
		{
         case 0:
			 inFileName=argv[i];
			 break;
         case 1:
			 threshold = atof(argv[i]);
			 THRESHOLD = threshold;
			 break;
         case 2:
			 outFileName=argv[i];
			 break;
		 case 3:
			 nofremovedT = atoi(argv[i]);
			 break;
		 case 4:
			 startPos = atoi(argv[i]);
		}
		i++;
    }

	clock_t startTime=clock();

	Data* fdat=new Data(inFileName);

	char tempFileName[128];
	strcpy(tempFileName,inFileName);
	strcat(tempFileName,".close");

	Data* fdatClose = new Data(tempFileName);

	if(!fdat->isOpen()) {
		cerr << inFileName << " could not be opened!" << endl;
		exit(2);
	}

	fp_buf=new memory(60, 4194304L, 8388608L, 2);
	//hash_buf=new memory(60, 4194304L, 8388608L, 2);

	patternTable=new hash_table();
	patternTable->readClose(fdatClose);

	T_tree* ttree = (T_tree*) fp_buf->newbuf(1, sizeof(T_tree));
	ttree->init();

	printf("Reading inputs...");

	ttree->scan_DB(fdat);

	printf("Input reading time: %f\n", (clock()-startTime)*1.0/CLOCKS_PER_SEC);

	printf("Updating...");
	//scanf("hold");

	clock_t start =  clock();
	int frontend=0;
	int *tset = new int [nofremovedT];

	if(startPos>=TRANSACTION_NO)
		startPos -= TRANSACTION_NO;

	if(startPos+nofremovedT>TRANSACTION_NO)
	{
		frontend = startPos+nofremovedT - TRANSACTION_NO; //rewind it
		for(i=0;i<frontend;i++)
			tset[i] = i;

		int j=0;
		for(;i<nofremovedT;i++)
		{
			tset[i] = j+startPos;
			j++;
			if(tset[i]>=TRANSACTION_NO)
			{
				printf("remove trans error!!");
				scanf("exit");
				exit(1);
			}
		}
	}
	else
	{
		for(i=0; i<nofremovedT; i++)
			tset[i]=i+startPos;
	}

	ttree->removeTrans(tset, nofremovedT);
	//tset[0]= 3;
	//ttree->removeTrans(tset,1);
	TRANSACTION_NO = TRANSACTION_NO-nofremovedT;

	delete []tset;

	FSout *fout = new FSout(outFileName);
	FSout *foutClose=NULL;

	//char tempFileName[128];
	strcpy(tempFileName,outFileName);
	strcat(tempFileName,".close");
	foutClose = new FSout(tempFileName);


	double closePatterns=0;
	double keyPatterns=0;

	printf("Total update time: %f\n", (clock()-start)*1.0/CLOCKS_PER_SEC);
	printf("Output resutls...");

	clock_t start_p =  clock();
	//foutClose->printTime("Total Time",(clock()-startTime)*1.0/CLOCKS_PER_SEC);
	foutClose->printTime("Total update time: ", (clock()-start)*1.0/CLOCKS_PER_SEC);
	foutClose->printString("\n");

	if(fullNum>0)
	{
		foutClose->printOrderSet(fullNum, fullItem, TRANSACTION_NO);
		foutClose->printString("\n");
	}


	patternTable->outputClose(foutClose, keyPatterns, closePatterns);
	foutClose->close();
	delete foutClose;

	ttree->outputTree(fout);
	fout->close();
	delete fout;

	//printf("Output time: %f\n", (clock()-start_p)*1.0/CLOCKS_PER_SEC);

	delete fdat;
	delete fp_buf; //added to avoid memory leak
	delete patternTable;

	clock_t finishTime=clock();
	double execTime=(finishTime-startTime)*1.0/CLOCKS_PER_SEC;
	if(threshold>1)
		threshold=threshold*1.0/TRANSACTION_NO;
	threshold=threshold*100;
	//printf("%s %s %0.2f %d -r %.1f, seconds: %.3f key patterns: %.0f close patterns: %.0f \n", "fpmax",inFileName,threshold,THRESHOLD, RATIO_THRESHOLD,execTime,keyPatterns,closePatterns);

}
